
from mpi4py import MPI
from typing import Callable, Any
import numpy as np
import os

from my_.data.templates import gridded_data
from my_.files.netcdf import nc_open, variables_to_dict
from my_.files.handy import check_file_exists

def pixel_wise(func: Callable,
               variables: list[str],
               variables_out: list[str],
               data: dict | None = None,
               files: None | str = None,
               file_out: None | str = None,
               dtype: str = 'float32',
               year_start: int | None = None,
               year_end: int | None = None,
               return_shape: int | list[int] = 1,
               return_dims: str | list[str] = 'time',
               *args, **kwargs) -> Any:
    
    if check_file_exists(file_out): return

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()
    
    print(f'Rank {rank} out of {size} active.')

    comm.Barrier()

    if rank == 0:

        print(f'Rank {rank} is loading the data...')
        
        if (data is None) and (files is None):
            NotImplementedError('No dataset supplied...')
        
        if data is not None:

            data_ = gridded_data(**data)

            arrays = data_.get_values(variables_ = variables,
                                      y0 = year_start,
                                      y1 = year_end,
                                      dtype = dtype)
            
        if files is not None:

            data_ = nc_open(files)

            arrays = variables_to_dict(data = data_,
                                       variables = variables,
                                       dtype = dtype)

        shapes = {v: arrays[v].shape for v in variables}

        if isinstance(return_shape, int): return_shape = [return_shape]

        arrays_out = {v: np.empty([*return_shape, *shapes[variables[0]][-2:]],
                                   dtype = dtype) 
                                   for v in variables_out}
        
        for v in variables_out: arrays_out[v][:] = np.nan     

    elif rank != 0:

        print(f'Rank {rank} creating empty objects...')

        shapes = None

    shapes = comm.bcast(shapes, root = 0)
                                   

    if rank != 0:

        print(f'I am rank {rank} and I know the shapes: {shapes}.')
        print('Creating empty arrays...')

        arrays_recv = {v: np.empty(shapes[v][:-2], 
                                   dtype = dtype) 
                                   for v in variables}

                                   
    comm.Barrier()


    print(f'I am rank {rank}. Ready for sending / receiving and calculations.')

    range_y = np.arange(shapes[variables[0]][-2])
    range_x = np.arange(shapes[variables[0]][-1])
    grid = np.array(np.meshgrid(range_y, range_x))
    cells = grid.T.reshape(-1, 2)
    length = len(cells)
    perrank = length // (size - 1)
    resid = length - perrank * (size - 1)

    for yx in range(perrank):
    
        if rank == 0:

            for r in range(1, size):

                r_n = (size - 1) * yx + (r - 1)
                y = cells[r_n][0]
                x = cells[r_n][1]

                arrays_yx = {v: arrays[v][..., y, x] for v in variables}

                print(f'Rank {rank} sending cell {r_n} corresponding to')
                print(f'grid indices {x}, {y} to rank {r}.')

                for v, array in arrays_yx.items(): 

                    comm.Send(np.ascontiguousarray(array), 
                              dest = r, 
                              tag = yx)

                print(f'Rank {rank} succesfully sent data to rank {r}.')
            
            for r in range(1, size):

                r_n = (size - 1) * yx + (r - 1)
                y = cells[r_n][0]
                x = cells[r_n][1]

                print(f'Rank {rank} receiving func {func.__name__} output')
                print(f'from grid indices {x}, {y} to rank {r}.')

                return_object = comm.recv(source = r, tag = yx)

                print(f'Rank {rank} received function output from rank {r},')
                print(f'from grid indices {x}, {y} to rank {r}.')

                for v in variables_out:

                    arrays_out[v][..., y, x] = return_object[v]

                    
        if rank != 0:

            r_m = (size - 1) * yx + (rank - 1)
            y = cells[r_m][0]
            x = cells[r_m][1]

            print(f'Rank {rank} receiving cell {r_m} corresponding to')
            print(f'grid indices {x}, {y} from rank 0.')

            for v in variables: comm.Recv(arrays_recv[v], source = 0, tag = yx)

            print(f'Rank {rank} received cell {r_m}.')
            print(f'Now executing function {func.__name__}...')

            out_func = func(arrays_recv, *args, **kwargs)

            print(f'Rank {rank} executed function for grid indices {x}, {y}.')
            print(f'Sending back the output to rank 0...')

            comm.send(out_func, dest = 0, tag = yx)
            
            print(f'Rank {rank} succesfully sent function output')
            print(f'for grid indices {x}, {y} to rank 0.')
    

    comm.Barrier()


    if (resid != 0) and (rank <= resid):

        print(f'Rank {rank} starting residual loop.')

        if rank == 0:
        
            for r in range(1, resid + 1):

                y = cells[-r][0]
                x = cells[-r][1]

                arrays_yx = {v: arrays[v][..., y, x] for v in variables}

                print(f'Rank {rank} sending cell {-r} corresponding to')
                print(f'grid indices {x}, {y} to rank {r}.')

                for v, array in arrays_yx.items(): 

                    comm.Send(np.ascontiguousarray(array),
                              dest = r, 
                              tag = yx)

                print(f'Rank {rank} succesfully sent data to rank {r}.')

            for r in range(1, resid + 1):

                y = cells[-r][0]
                x = cells[-r][1]

                print(f'Rank {rank} receiving func {func.__name__} output')
                print(f'from grid indices {x}, {y} to rank {r}.')

                return_object = comm.recv(source = r, tag = yx)

                print(f'Rank {rank} received function output from rank {r},')
                print(f'from grid indices {x}, {y} to rank {r}.')

                for v in variables_out:

                    arrays_out[v][..., y, x] = return_object[v]


        if rank != 0:

            y = cells[-rank][0]
            x = cells[-rank][1]

            print(f'Rank {rank} receiving cell {-rank} corresponding to')
            print(f'grid indices {x}, {y} from rank 0.')
            
            for v in variables: comm.Recv(arrays_recv[v], source = 0, tag = yx)

            print(f'Rank {rank} received cell {-rank}.')
            print(f'Now executing function {func.__name__}...')

            out_func = func(arrays_recv, *args, **kwargs)

            print(f'Rank {rank} executed function for grid indices {x}, {y}.')
            print(f'Sending back the output to rank 0...')

            comm.send(out_func, dest = 0, tag = yx)
            
            print(f'Rank {rank} succesfully sent function output')
            print(f'for grid indices {x}, {y} to rank 0.')


    comm.Barrier()


    if rank == 0:


        if file_out is not None:

            import xarray as xr

            if isinstance(return_dims, str): return_dims = [return_dims]

            out_vars = {v: ([*return_dims, 'lat', 'lon'], arrays_out[v]) for v in variables_out}

            DS_out = xr.Dataset(data_vars = out_vars)

            print('Now wrinting DS to netcdf...')

            DS_out.to_netcdf(path = file_out, mode='w')

            print('Script was successful! Bye bye!')
    
        else:
    
            print('Script was successful! Bye bye!')
    
            return arrays_out


def along_dim(func: Callable,
              dim: int,
              variables: list[str],
              variables_out: list[str],
              data: dict | None = None,
              files: str | None = None,
              file_out: None | str = None,
              dtype: str = 'float32',
              year_start: int | None = None,
              year_end: int | None = None,
              return_dims: str | list[str] = 'time',
              *args, **kwargs) -> Any:

    if check_file_exists(file_out): return

    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()
    
    print(f'Rank {rank} out of {size} active.')

    comm.Barrier()

    if rank == 0:

        print(f'Rank {rank} is loading the data...')
        
        if (data is None) and (files is None):
            NotImplementedError('No dataset supplied...')
        
        if data is not None:

            data_ = gridded_data(**data)

            arrays = data_.get_values(variables_ = variables,
                                      y0 = year_start,
                                      y1 = year_end,
                                      dtype = dtype)
            
        if files is not None:

            data_ = nc_open(files)

            arrays = variables_to_dict(data = data_,
                                       variables = variables,
                                       dtype = dtype)

        shapes = {v: arrays[v].shape for v in variables}

        arrays_out = {v: np.empty([*shapes[v]],
                                   dtype = dtype) 
                                   for v in variables_out}
        
        for v in variables_out: arrays_out[v][:] = np.nan     

    elif rank != 0:

        print(f'Rank {rank} creating empty objects...')

        shapes = None

    shapes = comm.bcast(shapes, root = 0)
                                   

    if rank != 0:

        print(f'I am rank {rank} and I know the shapes: {shapes}.')
        print('Creating empty arrays...')

        arrays_recv = {v: np.empty(shapes[v][:dim] +  
                                   shapes[v][dim + 1:],
                                   dtype = dtype) 
                                   for v in variables}

                                   
    comm.Barrier()


    print(f'I am rank {rank}. Ready for sending / receiving and calculations.')

    length = shapes[variables[0]][dim]
    ndims = len(shapes[variables[0]])
    
    perrank = length // (size - 1)
    resid = length - perrank * (size - 1)

    for idx in range(perrank):
    
        if rank == 0:

            for r in range(1, size):

                r_n = (size - 1) * idx + (r - 1)
                
                slx = [slice(0, None)] * ndims
                slx[dim] = r_n
                slx = tuple(slx)

                arrays_yx = {v: arrays[v][slx] for v in variables}

                print(f'Rank {rank} sending cell {r_n} corresponding to')
                print(f'dimension {dim}, index {r_n} to rank {r}.')

                for v, array in arrays_yx.items(): 

                    comm.Send(np.ascontiguousarray(array), 
                              dest = r, 
                              tag = idx)

                print(f'Rank {rank} succesfully sent data to rank {r}.')
            
            for r in range(1, size):

                r_n = (size - 1) * idx + (r - 1)

                slx = [slice(0, None)] * ndims
                slx[dim] = r_n
                slx = tuple(slx)

                print(f'Rank {rank} receiving func {func.__name__} output')
                print(f'dimension {dim}, index {r_n} from rank {r}.')

                return_object = comm.recv(source = r, tag = idx)

                print(f'Rank {rank} received function output from rank {r},')
                print(f'dimension {dim}, index {r_n} from rank {r}.')

                for v in variables_out:

                    arrays_out[v][slx] = return_object[v]

                    
        if rank != 0:

            r_m = (size - 1) * idx + (rank - 1)

            print(f'Rank {rank} receiving cell {r_m} corresponding to')
            print(f'dimension {dim}, index {r_m} from rank 0.')

            for v in variables: comm.Recv(arrays_recv[v], source = 0, tag = idx)

            print(f'Rank {rank} received cell {r_m}.')
            print(f'Now executing function {func.__name__}...')

            out_func = func(arrays_recv, *args, **kwargs)

            print(f'Rank {rank} executed function for dimension {dim}, index {r_m}.')
            print(f'Sending back the output to rank 0...')

            comm.send(out_func, dest = 0, tag = idx)
            
            print(f'Rank {rank} succesfully sent function output')
            print(f'dimension {dim}, index {r_m} to rank 0.')
    

    comm.Barrier()

    if (resid != 0) and (rank <= resid):

        print(f'Rank {rank} starting residual loop.')

        if rank == 0:
        
            for r in range(1, resid + 1):
                
                slx = [slice(0, None)] * ndims
                slx[dim] = -r
                slx = tuple(slx)

                arrays_yx = {v: arrays[v][slx] for v in variables}

                print(f'Rank {rank} sending cell {-r} corresponding to')
                print(f'dimension {dim}, index {-r} to rank {r}.')

                for v, array in arrays_yx.items(): 

                    comm.Send(np.ascontiguousarray(array),
                              dest = r, 
                              tag = r)

                print(f'Rank {rank} succesfully sent data to rank {r}.')

            for r in range(1, resid + 1):

                slx = [slice(0, None)] * ndims
                slx[dim] = -r
                slx = tuple(slx)

                print(f'Rank {rank} receiving func {func.__name__} output')
                print(f'dimension {dim}, index {-r} from rank {r}.')

                return_object = comm.recv(source = r, tag = r)

                print(f'Rank {rank} received function output from rank {r},')
                print(f'dimension {dim}, index {-r} from rank {r}.')

                for v in variables_out:

                    arrays_out[v][slx] = return_object[v]


        if rank != 0:

            print(f'Rank {rank} receiving cell {-rank} corresponding to')
            print(f'dimension {dim}, index {-rank} from rank 0.')
            
            for v in variables: comm.Recv(arrays_recv[v], source = 0, tag = rank)

            print(f'Rank {rank} received cell {-rank}.')
            print(f'Now executing function {func.__name__}...')

            out_func = func(arrays_recv, *args, **kwargs)

            print(f'Rank {rank} executed function for dimension {dim}, index {-rank}.')
            print(f'Sending back the output to rank 0...')

            comm.send(out_func, dest = 0, tag = rank)
            
            print(f'Rank {rank} succesfully sent function output')
            print(f'for dimension {dim}, index {-rank} to rank 0.')


    comm.Barrier()


    if rank == 0:


        if file_out is not None:

            import xarray as xr

            if isinstance(return_dims, str): return_dims = [return_dims]

            out_vars = {v: ([*return_dims, 'lat', 'lon'], arrays_out[v]) for v in variables_out}

            DS_out = xr.Dataset(data_vars = out_vars)

            print('Now wrinting DS to netcdf...')

            DS_out.to_netcdf(path = file_out, mode = 'w', format = 'NETCDF4_CLASSIC')

            print('Script was successful! Bye bye!')
    
        else:
    
            print('Script was successful! Bye bye!')
    
            return arrays_out




if __name__ == '__main__':

    from my_.science.anomalies import standard_index

    pixel_wise(func = standard_index,
               variables = ['QFLX_EVAP_GRND'],
               variables_out = ['SXI_ET_GRND'],
               files =  '/p/scratch/cjibg31/jibg3105/data/CLM5EU3/006/join_8d/*.nc',
               file_out = 'SXI_ET_GRND.nc',
               return_shape = 46*24,
               return_dims = 'time',
               year_start = 1995,
               year_end = 1997)



import deims
import pandas as pd
from datarie.handy import create_dirs, check_file_exists, save_df
from datarie.csv import open_csv


def from_network(network: str = 'eLTER',
                 verified: bool = True) -> list:
    
    network_codes = {'eLTER': '4742ffca-65ac-4aae-815f-83738500a1fc',
                     'DLTER': 'e904354a-f3a0-40ce-a9b5-61741f66c824',
                     'ICOS': '80633d38-4c85-4ee0-a4ce-7bbbd99c888c'}
    

    sites = deims.getListOfSites(network = network_codes[network], 
                                           verified_only = verified)

    return sites


def information(id: str):

    information     = deims.getSiteById(id)

    return information


def coordinates(id: str):

    coordinates = information(id)['attributes']['geographic']['coordinates']

    latitude = float(coordinates.replace('(', '').replace(')', '').split(' ')[2])
    longitude = float(coordinates.replace('(', '').replace(')', '').split(' ')[1])

    return latitude, longitude


def ecosystem_type(id: str):

    ecosystem_type  = information(id)['attributes']['environmentalCharacteristics']['eunisHabitat']
    
    if ecosystem_type is None: return 'None'

    return ecosystem_type[0]['label']


def name(id: str):

    name            = information(id)['title']

    return name


def sites_df(network: str = 'eLTER',
             verified: bool = True,
             out_dir: str = 'deims',
             out_file: str = 'sites.csv'):

    out = f'{out_dir}/{out_file}'

    if check_file_exists(out): 
        
        print(f'\n{out} file already exists!\n')
        
        return open_csv(out, index_col = 0)
    
    print(f'\nSaving site file {out}.\n')

    create_dirs(out_dir)

    ids = from_network(network,
                       verified)

    names = [name(s) for s in ids]

    coords = [coordinates(s) for s in ids]
    latitudes = [c[0] for c in coords]
    longitudes = [c[1] for c in coords] 
    ecosystem_types = [ecosystem_type(s) for s in ids]

    df = pd.DataFrame({'name': names,
                       'latitude': latitudes,
                       'longitude': longitudes,
                       'ecosystem_type': ecosystem_types},
                       index = ids)

    save_df(df, out)

    return df




    


import pandas as pd
import re
import xarray as xr

def index(y0: int, 
          y1: int, 
          t_res: str,
          month0: int = 1,
          month1: int = 12, 
          day0: int = 1,
          day1: int = 31,
          leapday: bool = True,
          range_kwargs: dict = {},
          resample_kwargs: dict = {}) -> pd.Series:

    """
    Create time index from beginning of year 0
    to end of year 0;
    To use for pandas time series;
    """
    
    if t_res[-1] == 'H': end_time = str(24 - int(t_res[:-1]))
    else: end_time = '23:59:00'

    y0_time = f'{y0}-{month0:02}-{day0:02} 00:00:00'
    y1_time = f'{y1}-{month1:02}-{day1:02} {end_time}'

    t_res_comp = re.split('(\d+)', t_res)

    t_res_str = t_res_comp[-1]

    time_raw = pd.date_range(y0_time, 
                             y1_time, 
                             freq = t_res_str,
                             **range_kwargs).to_series()

    # Make yearly groups for resample
    time_groups = time_raw.groupby(time_raw.dt.year, 
                                   group_keys=False)

    # Resample. Not always needed (but e.g. for multiples 
    # '8D' frequecies)
    time_resampled = time_groups.resample(t_res, 
                                          **resample_kwargs) \
                                          .asfreq()
    
    if leapday == False: 
        time_resampled = drop_leapday(time_resampled)

    return time_resampled


def drop_leapday(series):

    """
    Drop the leapday entries from any
    time series. I think there should be
    an easier way as long as index is 
    datetime.
    """

    mask = [False if (s.month == 2 and s.day == 29) else True for s in series.index]

    series_out = series[mask]

    return series_out


def xr_resample(df: xr.Dataset | xr.DataArray,
                offset_str: str, 
                method: str = 'mean',
                **kwargs) -> xr.Dataset | xr.DataArray:
    
    func_ = getattr(xr.DataArray, method)
        
    df_resampler = df.resample(time = offset_str)

    df_resampled = func_(df_resampler, **kwargs)

    return df_resampled


def pd_resample(df: pd.DataFrame | pd.Series,
                offset_str: str, 
                method: str = 'mean',
                yearly_agg: bool = True) -> pd.DataFrame | pd.Series:
    
    print(f'\nResampling dataframe by {method} to {offset_str} time frequency...\n')
    
    index = pd.to_datetime(df.index).year

    if yearly_agg: df_n = df.groupby(index, group_keys = False)

    df_resampler = df_n.resample(offset_str)

    df_resampled = df_resampler.agg(method)

    return df_resampled



import pandas as pd
import numpy as np
import scipy.stats as stats
import uncertainties.unumpy as unp
import uncertainties as unc
from scipy.optimize import curve_fit
from typing import Callable
    
def mean(array, axis: int | tuple | None = None, std_error = False):

    array_mean = np.nanmean(array, axis = axis)

    return array_mean


def std_error_mean(array, axis: int | tuple | None = None):

    array_std = np.nanstd(array, axis = axis)

    array_n = np.count_nonzero(~np.isnan(array), axis = axis)

    array_std_error_mean= array_std / np.sqrt(array_n)

    return array_std_error_mean


def variance(array, axis: int | tuple | None = None):

    array_var = np.nanvar(array, axis = axis)

    return array_var


def std_deviation(array, axis: int | tuple | None = None):

    array_std = np.nanstd(array, axis = axis)

    return array_std


def std_error_variance(array, axis: int | tuple | None = None):

    array_var = np.nanvar(array, axis = axis)

    array_n = np.count_nonzero(~np.isnan(array), axis = axis)

    array_std_error_var = np.sqrt(2 / (array_n - 1)) * array_var

    return array_std_error_var


def skewness(array, axis: int | None = None, nan_policy = 'omit'):

    from scipy.stats import skew

    array_skew = skew(array, axis, nan_policy = nan_policy)

    return array_skew


def std_error_skewness(array, axis: int | tuple | None = None):

    array_n = np.count_nonzero(~np.isnan(array), axis = axis)

    array_std_error_sk = np.sqrt((6 * array_n * (array_n - 1)) / 
                                  ((array_n - 2) * (array_n + 1) * (array_n + 3)))

    return array_std_error_sk


def kurtosis(array, axis: int | None = None, nan_policy = 'omit'):

    from scipy.stats import kurtosis

    array_kurtosis = kurtosis(array, axis, nan_policy = nan_policy)

    return array_kurtosis


def std_error_kurtosis(array, axis: int | tuple | None = None):

    array_n = np.count_nonzero(~np.isnan(array), axis = axis)

    array_std_error_sk = std_error_skewness(array, axis = axis)
    
    array_std_error_krt = 2 * array_std_error_sk * np.sqrt((array_n**2 - 1) / ((array_n - 3) * (array_n + 5)))

    return array_std_error_krt


def prediction_band(x: np.ndarray,
                    y: np.ndarray,
                    level: float = 0.95,
                    n_pts: None | int = None):
    
    if n_pts is None: n_pts = len(x)

    mask = ~np.isnan(x) & ~np.isnan(y)

    x_clean, y_clean = x[mask], y[mask]

    lr = stats.linregress(x_clean, y_clean)

    trend = lr.intercept + x_clean * lr.slope

    tinv = lambda q, nu: abs(stats.t.ppf(q/2, nu))
    
    ts = tinv(1 - level, len(x_clean) - 2)

    slopes = np.linspace(lr.slope + ts * lr.stderr, 
                         lr.slope - ts * lr.stderr,
                         n_pts)
    
    intercepts = np.linspace(lr.intercept + ts * lr.intercept_stderr, 
                             lr.intercept - ts * lr.intercept_stderr, 
                             n_pts)
    
    lines = np.empty([n_pts, len(x_clean)])
    
    x_fill = np.sort(x_clean)

    for i in range(len(slopes)):

        lines[i,:] = x_fill * slopes[i] + intercepts[i]

    lower = np.min(lines, axis = 0)
    upper = np.max(lines, axis = 0)

    return x_clean, trend, lower, upper

def confidence_band(x: np.ndarray,
                    y: np.ndarray,
                    level: float = 0.95,
                    n_pts: None | int = None,
                    sides: int = 2,
                    f: Callable = lambda a, x, b: a * x + b):

    if n_pts is None: n_pts = len(x)

    mask = ~np.isnan(x) & ~np.isnan(y)
    x_clean, y_clean = x[mask], y[mask]

    popt, pcov = curve_fit(f, x_clean, y_clean)
    
    a, b = unc.correlated_values(popt, pcov)
    
    N, n = x_clean.size, len(popt)

    q = stats.t.ppf(1 - ((1 - level) / sides), N - n)
    
    py = a * x_clean + b

    nom = unp.nominal_values(py)
    std = unp.std_devs(py)

    lower = nom - q * std 
    upper = nom + q * std

    return x_clean, nom, lower, upper





import numpy as np
import os

from sklearn.cluster import KMeans, DBSCAN, OPTICS
from sklearn.metrics import silhouette_samples, silhouette_score

def kmeans(data: list[np.ndarray],
           n_clusters: int | list[int]):
    
    if isinstance(n_clusters, int): n_clusters = [n_clusters]#

    print(f'\nCompute kmeans clusters (N = {n_clusters})...\n')
    
    data_stacked = np.stack(data, axis = -1)
    data_valid = ~np.isnan(data_stacked)

    mask = np.all(data_valid, axis = -1)

    data_masked = data_stacked[mask]
    
    labels = []
    centroids = []
    scores = []

    for i, nc in enumerate(n_clusters):

        kmeans = KMeans(n_clusters = nc)
        labels.append(kmeans.fit_predict(data_masked))
        centroids.append(kmeans.cluster_centers_)

        scores.append(silhouette_score(data_masked, 
                                        labels[i]))
            
        print(f'\nFor n_clusters = {nc}, silhouette score is {scores[i]}...\n')
    
    i_opt = np.argmax(scores)
    
    labels_opt = labels[i_opt]
    centroids_opt = centroids[i_opt]
    scores_opt = scores[i_opt]

    return labels_opt, centroids_opt, scores_opt


def dbscan(data: list[np.ndarray],
           weights: np.ndarray | None = None,
           **kwargs):
    
    from joblib import effective_n_jobs

    print(f'\nCompute DBscan clusters')
    print(f'with {effective_n_jobs(-1)} cpus...\n')

    shape = data[0].shape

    if weights is None: weights = np.ones(shape)

    data.append(weights)

    data_stacked = np.stack(data, 
                            axis = -1)
    
    data_valid = ~np.isnan(data_stacked)

    mask = np.all(data_valid, axis = -1)

    data_masked = data_stacked[mask]

    print(f'\nNumber of points:')
    print(f'{data_masked.shape[0]}\n')
    print(f'\nNumber of features:') 
    print(f'{data_masked.shape[1]-1}\n')

    clusterer = DBSCAN(**kwargs, n_jobs = -1)
        
    labels = clusterer.fit_predict(data_masked[:, :-1],
                                   sample_weight = data_masked[:, -1])
    
    array_r = np.zeros(shape)
    array_r[:] = np.nan
    array_r[mask] = labels

    return array_r


def optics(data: list[np.ndarray],
           **kwargs):
    
    from joblib import effective_n_jobs, parallel_config
    from dask_jobqueue import SLURMCluster
    from dask.distributed import Client
    
    #cluster = SLURMCluster(processes = 1, cores = 128, memory = '1GB')
    #client = Client(cluster)

    print(f'\nCompute DBscan clusters')
    print(f'with {effective_n_jobs(-1)} cpus...\n')

    shape = data[0].shape
    
    data_stacked = np.stack(data, axis = -1)
    data_valid = ~np.isnan(data_stacked)

    mask = np.all(data_valid, axis = -1)

    data_masked = data_stacked[mask]

    print(f'\nNumber of points:')
    print(f'{data_masked.shape[0]}\n')
    print(f'\nNumber of features:') 
    print(f'{data_masked.shape[1]}\n')

    clusterer = OPTICS(**kwargs, n_jobs = -1)

    #with parallel_config('dask', n_jobs = -1):
        
    labels = clusterer.fit_predict(data_masked)
    
    array_r = np.zeros(shape)
    array_r[:] = np.nan
    array_r[mask] = labels

    return array_r


def hp_dbscan(file: os.PathLike,
              dataset: str,
              dir_build: os.PathLike,
              epsilon: float,
              min_points: int):
    
    import subprocess
    #import sys
    #sys.path.append(dir_build)
    #
    #import hpdbscan
    #import h5py

    #print(f'\nCompute DBSCAN clusters')
    #print(f'in highly parallel fashion...\n')

    #clusterer = hpdbscan.HPDBSCAN(epsilon = epsilon,
    #                              min_points = min_points)
    #
    #labels = clusterer.cluster(file, dataset)


    command = ' '.join([
              f'srun',
              '-n', '128',
              f'{dir_build}/hpdbscan',
              '-t', '128',
              '-m', f'{min_points}',
              '-e', f'{epsilon}',
              '-i', f'{file}',
              '-o', f'{file}',
              '--input-dataset',
              f'{dataset}',
              '--output-dataset',
              f'CLUSTERS'])

    print('\nExecute:')
    print(f'\n{command}\n')

    subprocess.call(command, shell = True)

    print(f'\nDone!\n')



import os

from glob import glob
import xarray as xr
import numpy as np
import pandas as pd
import netCDF4 as nc
import timeit

from my_.data.templates import gridded_data
from my_.files.handy import create_dirs, check_file_exists

past_global_daily = {
    'name': 'MSWEP_past_global_daily',
    'version': (1, 0, 6),
    'path': '/p/scratch/cjibg31/jibg3105/data/GLEAM/MSWEP/yearly/',
    'type_file': 'netcdf',
    'year_start': 1979,
    'month_start': 1,
    'year_end': 2020,
    'month_end': 12,
    'resolution_time': 'D',
    'grid': 'MSWEP',
    'variables': ['P'],
    'variable_names': {'P': 'precipitation'},
    'variable_dimensions': {'P': ['time', 'lat', 'lon']}, 
    'variable_units': {'P': 'mm/day'},
    'mask_value': None,
    'leapday': True,}


def create_yearly_files(path_rawdata: os.PathLike[str]):
    
    print('\nConvert MSWEP files to yearly...\n')

    data_ = gridded_data(**past_global_daily)

    create_dirs(data_.path)

    time = data_.index_time()
    
    years = np.unique(time.dt.year.to_numpy())

    for y in years:

        start_time = timeit.default_timer()

        files = sorted(glob(f'{path_rawdata}/{y}*.nc'))

        data_y = xr.open_mfdataset(files, 
                                   combine = 'by_coords',
                                   chunks = {'time': len(files), 
                                             'lat': 32, 
                                             'lon': 32},
                                    engine = 'netcdf4').load()
        
        print(f'Create yearly file for year {y}...\n')

        if check_file_exists(f'{data_.path}/{y}.nc'): continue

        data_y.to_netcdf(f'{data_.path}/{y}.nc',
                         format = 'NETCDF4_CLASSIC', 
                         unlimited_dims = ['time'])
        
        data_y.close()

        print('Time for processing:')
        print(timeit.default_timer() - start_time)  
        print('Continue...')


import os
from datarie.handy import create_dirs, check_file_exists
from glob import glob
import xarray as xr
import numpy as np
import pandas as pd

NOAH025_all_EUROCORDEX_monthly = {
    'name': 'GLDAS_NOAH025_M',
    'version': (1, 0, 6),
    'path': '/p/scratch/cjibg31/jibg3105/data/GLDAS/NOAH025_M/',
    'type_file': 'netcdf',
    'year_start': 1980,
    'month_start': 1,
    'year_end': 2014,
    'month_end': 12,
    'resolution_time': 'MS',
    'grid': 'GLDAS',
    'variables': ['P',
                  'ET',
                  'PET',
                  'Runoff',
                  'SM'],
    'variable_names': {'P': 'Rainf_f_tavg',
                       'ET': 'Evap_tavg',
                       'PET': 'PotEvap_tavg',
                       'Runoff': 'ro',
                       'SM': 'RootMoist_inst'},
    'variable_dimensions': {'P': ['time', 'lat', 'lon'], 
                            'ET': ['time', 'lat', 'lon'],
                            'PET': ['time', 'lat', 'lon'],
                            'Runoff': ['time', 'lat', 'lon'],
                            'SM': ['time', 'layer', 'lat', 'lon']}, 
    'variable_units': {'P': 'm/day',
                       'ET': 'm/day',
                       'PET': 'm/day',
                       'Runoff': 'm/day',
                       'SM': 'm^3/m^3'},
    'mask_value': None,
    'leapday': True,}


def create_yearly_files(path_rawdata: str,
                        path_out: str):
    
    """
    Input directory should only contain the GLEAM rawdata files...
    """

    print('\nConvert GLDAS files to yearly...\n')
    
    create_dirs(path_out)

    files = glob(f'{path_rawdata}/*')

    data_raw = xr.open_mfdataset(files).load()
    
    years = np.unique(data_raw.time.dt.year.values)

    for y in years:

        print(f'Create yearly file for year {y}...\n')

        if check_file_exists(f'{path_out}/{y}.nc'): continue

        data_y = data_raw.sel(time = data_raw.time.dt.year.isin([y]))

        data_y.to_netcdf(f'{path_out}/{y}.nc',
                         format = 'NETCDF4_CLASSIC', 
                         unlimited_dims = ['time'])




def style_1(font_dir: str = '/p/scratch/cjibg31/jibg3105/projects/my_py/src/my_/plot/fonts'):

    import matplotlib as mpl
    import matplotlib.font_manager as fm


    
    font_files = fm.findSystemFonts(fontpaths = font_dir)
    
    for font_file in font_files:
        fm.fontManager.addfont(font_file)
    
    prop = fm.FontProperties(fname=font_dir+'/Montserrat-Medium.otf')

    mpl.rcParams['font.family']     = prop.get_name()
    text_color                      = 'dimgrey'
    mpl.rcParams['xtick.color']     = text_color
    mpl.rcParams['ytick.color']     = text_color


def nospines(ax):

    ax.spines.right.set_visible(False)
    ax.spines.top.set_visible(False)
    ax.spines.bottom.set_visible(False)
    ax.spines.left.set_visible(False)

    



import pandas as pd
import numpy as np
from my_.series.group import select_multi_index
    

def concat(dfs: list[pd.DataFrame], 
           join: str = 'outer', 
           axis: int = 1, 
           sort: bool = False, 
           remove_duplicates: bool = True):

    if axis == 1: dfs = add_missing_column_levels(dfs)

    df = pd.concat(dfs, join = join, axis = axis)

    if sort: df = df.sort_index(axis = axis)

    if remove_duplicates: df = df.loc[:, ~df.columns.duplicated()].copy()

    return df


def add_missing_column_levels(dfs: list[pd.DataFrame]):

    from my_.series.group import add_level_to_multi_columns
    
    nlevels                 = [df.columns.nlevels for df in dfs]

    if nlevels.count(nlevels[0]) == len(nlevels): return dfs

    max_levels              = max(nlevels)

    i_max_levels            = nlevels.index(max_levels)

    max_columns             = set(dfs[i_max_levels].columns.names)

    new_dfs                 = []

    for df in dfs:

        diff_columns        = set(max_columns).difference(set(df.columns.names))

        for missing_level in diff_columns:

            new_level       = ['-1'] * len(df.columns)

            new_loc         = len(df.columns.names)

            df              = add_level_to_multi_columns(df, new_loc, missing_level, new_level)

        new_dfs.append(df)

    return new_dfs

            

def mask_ge(df, masker, value, skip_mask_NA = True):

    import pandas as pd

    if skip_mask_NA:

        df_masked           = df.where(masker.values >= value, axis = 0)
    
    elif not skip_mask_NA:

        df_masked           = df.where(((masker.values >= value) | (masker.values == pd.NA)), axis = 0)

    return df_masked


def mask_le(df, masker, value, skip_mask_NA = True):

    import pandas as pd

    if skip_mask_NA:

        df_masked           = df.where(masker.values <= value, axis = 0)
    
    elif not skip_mask_NA:

        df_masked           = df.where(((masker.values <= value) | (masker.values == pd.NA)), axis = 0)

    return df_masked


def groupwise(df: pd.DataFrame,
              func) -> pd.DataFrame:
    
    #if isinstance(func, callable): func = [func]

    for f_ in func:

        df = f_(df)

    return df
    

def apply(df: pd.DataFrame, 
          method: str):

    # find the method in numpy attributes
    method_pd               = getattr(pd, method)

    df_new                  = df.agg(method_pd, skipna = False)

    return df_new


def column_wise(df, ffunc):

    import pandas as pd
    from itertools import product

    columns                 = df.columns

    df_out                  = pd.DataFrame(columns = columns, index = columns)

    iterator                = product(columns, columns)

    for col1, col2 in iterator:

        col1_v              = df[col1]

        col2_v              = df[col2]

        df_out.loc[col1, col2] = ffunc(col1_v, col2_v)

    return df_out


def single_axis_wise(df: pd.DataFrame, 
                     level: str, 
                     key: str, 
                     ffunc: callable, 
                     axis: int = 1,
                     ffunc_args: dict = {}):
    
    from my_.series.group import select_multi_index

    if axis == 1:

        axx = df.columns

    elif axis == 0:

        axx = df.index

    else: NotImplementedError

    keys = axx.get_level_values(level)

    index = axx.droplevel(level).unique()

    df_out = pd.DataFrame(columns = index,  index = keys)

    independent = select_multi_index(df, 
                                     axis = axis,
                                     levels = level, 
                                     keys = key)

    for ix, xx in enumerate(axx):

        dependent = df[xx]
        ix = keys[ix]
        
        df_out.loc[ix, index] = ffunc(independent, 
                                      dependent, 
                                      **ffunc_args)

    return df_out


def single_level_wise(df: pd.DataFrame,
                      level: str, 
                      key: str, 
                      ffunc: callable, 
                      axis: int = 1,
                      ffunc_args: dict = {}):

    if axis == 1:
        
        axx = df.columns

    elif axis == 0:
        
        axx = df.index
    
    else: NotImplementedError

    keys = axx.get_level_values(level).unique()

    df_out = pd.DataFrame(columns = [ffunc.__name__],  
                          index = keys)

    independent = select_multi_index(df, 
                                     levels = level, 
                                     keys = key,
                                     axis = axis)\
                                     .to_numpy()\
                                     .flatten()

    for value in keys:

        dependent = select_multi_index(df, 
                                       levels = level, 
                                       keys = value,
                                       axis = axis)\
                                       .to_numpy()\
                                       .flatten()

        df_out.loc[value, 
                   ffunc.__name__] = ffunc(independent, 
                                           dependent, 
                                           **ffunc_args)
    
    return df_out


def count_nonzero(df):

    import numpy as np

    df_np                   = df.to_numpy()

    return np.count_nonzero(~np.isnan(df_np))/6


pft = {

    'leafn': {
        'dim': 'pft',
        'ind': 2,
        'long_name': 'leaf N'
    },
    
    'leafn_storage': {
        'dim': 'pft',
        'ind': 3,
        'long_name': 'leaf N storage'
    },

    'leafn_xfer': {
        'dim': 'pft',
        'ind': 4,
        'long_name': 'leaf N transfer'
    },

    'frootn': {
        'dim': 'pft',
        'ind': 5,
        'long_name': 'fine root N'
    },

    'frootn_storage': {
        'dim': 'pft',
        'ind': 6,
        'long_name': 'fine root N storage'
    },

    'frootn_xfer': {
        'dim': 'pft',
        'ind': 7,
        'long_name': 'fine root N transfer'
    },

    'livestemn': {
        'dim': 'pft',
        'ind': 8,
        'long_name': 'live root N'
    },

    'livestemn_storage': {
        'dim': 'pft',
        'ind': 9,
        'long_name': 'live root N storage'
    },

    'livestemn_xfer': {
        'dim': 'pft',
        'ind': 10,
        'long_name': 'live root N transfer'
    },

    'deadstemn': {
        'dim': 'pft',
        'ind': 11,
        'long_name': 'dead stem N'
    },

    'deadstemn_storage': {
        'dim': 'pft',
        'ind': 12,
        'long_name': 'dead stem N storage'
    },

    'deadstemn_xfer': {
        'dim': 'pft',
        'ind': 13,
        'long_name': 'dead stem N transfer'
    },

    'livecrootn': {
        'dim': 'pft',
        'ind': 14,
        'long_name': 'live root N'
    },

    'livecrootn_storage': {
        'dim': 'pft',
        'ind': 15,
        'long_name': 'live root N storage'
    },

    'livecrootn_xfer': {
        'dim': 'pft',
        'ind': 16,
        'long_name': 'live root N transfer'
    },

    'deadcrootn': {
        'dim': 'pft',
        'ind': 17,
        'long_name': 'dead root N'
    },

    'deadcrootn_storage': {
        'dim': 'pft',
        'ind': 18,
        'long_name': 'dead root N storage'
    },

    'deadcrootn_xfer': {
        'dim': 'pft',
        'ind': 19,
        'long_name': 'dead root N transfer'
    },

    'npool': {
        'dim': 'pft',
        'ind': 20,
        'long_name': 'temporary plant N pool'
    },

    'leafc': {
        'dim': 'pft',
        'ind': 25,
        'long_name': 'leaf carbon'
    },

    'leafc_storage': {
        'dim': 'pft',
        'ind': 26,
        'long_name': 'leafc storage'
    },

    'leafc_xfer': {
        'dim': 'pft',
        'ind': 27,
        'long_name': 'leafc transfer'
    },

    'frootc': {
        'dim': 'pft',
        'ind': 29,
        'long_name': 'fine root C'    
    },

    'frootc_storage': {
        'dim': 'pft',
        'ind': 30,
        'long_name': ''    
    },

    'frootc_xfer': {
        'dim': 'pft',
        'ind': 31,
        'long_name': ''    
    },

    'livestemc': {
        'dim': 'pft',
        'ind': 32,
        'long_name': ''    
    },

    'livestemc_storage': {
        'dim': 'pft',
        'ind': 33,
        'long_name': ''    
    },

    'livestemc_xfer': {
        'dim': 'pft',
        'ind': 34,
        'long_name': ''    
    },

    'deadstemc': {
        'dim': 'pft',
        'ind': 35,
        'long_name': ''    
    },

    'deadstemc_storage': {
        'dim': 'pft',
        'ind': 36,
        'long_name': ''    
    },

    'deadstemc_xfer': {
        'dim': 'pft',
        'ind': 37,
        'long_name': ''    
    },

    'livecrootc': {
        'dim': 'pft',
        'ind': 38,
        'long_name': ''    
    },

    'livecrootc_storage': {
        'dim': 'pft',
        'ind': 39,
        'long_name': ''    
    },

    'livecrootc_xfer': {
        'dim': 'pft',
        'ind': 40,
        'long_name': ''    
    },

    'deadcrootc': {
        'dim': 'pft',
        'ind': 41,
        'long_name': ''    
    },

    'deadcrootc_storage': {
        'dim': 'pft',
        'ind': 42,
        'long_name': ''    
    },

    'deadcrootc_xfer': {
        'dim': 'pft',
        'ind': 43,
        'long_name': ''    
    },

    'gresp_storage': {
        'dim': 'pft',
        'ind': 44,
        'long_name': 'growth respiration storage'    
    },

    'gresp_xfer': {
        'dim': 'pft',
        'ind': 45,
        'long_name': ''    
    },

    'cpool': {
        'dim': 'pft',
        'ind': 46,
        'long_name': 'temporary photosynthate C pool'    
    },

    'totvegc': {
        'dim': 'pft',
        'ind': 48,
        'long_name': 'total vegetation carbon, excluding cpool'
    },

    'availc': {
        'dim': 'pft',
        'ind': 49,
        'long_name': 'C flux available for allocation'
    },

    'excess_cflux': {
        'dim': 'pft',
        'ind': 51,
        'long_name': 'C flux not allocated due to downregulation'
    }

}



soil = {

    'soil1n_vr': {
        'dim': ['column', 'levsoi'],
        'ind': 52,
        'long_name': 'SOIL1 (fast pool) N (vertically resolved)'
    },

    'soil2n_vr': {
        'dim': ['column', 'levsoi'],
        'ind': 53,
        'long_name': 'SOIL2 (medium pool) N (vertically resolved)'
    },

    'soil3n_vr': {
        'dim': ['column', 'levsoi'],
        'ind': 54,
        'long_name': 'SOIL3 (medium pool) N (vertically resolved)'
    },

    'soil1c_vr': {
        'dim': ['column', 'levsoi'],
        'ind': 55,
        'long_name': 'SOIL1 (fast pool) C (vertically resolved)'
    },

    'soil2c_vr': {
        'dim': ['column', 'levsoi'],
        'ind': 56,
        'long_name': 'SOIL2 (medium pool) C (vertically resolved)'
    },

    'soil3c_vr': {
        'dim': ['column', 'levsoi'],
        'ind': 57,
        'long_name': 'SOIL3 (medium pool) C (vertically resolved)'
    },
}


import pandas as pd
import pint
import numpy as np

def convert_units(data: pd. Series | pd.DataFrame | np.ndarray,
                  src_unit: str,
                  dst_unit: str):
        
        data_out = data.copy()
        
        if isinstance(data, pd.DataFrame): data = data.values
        if isinstance(data, pd.Series): data = data.values
        
        ureg = pint.UnitRegistry()
        Q_ = ureg.Quantity

        print(f'\nConverting units from {src_unit} to {dst_unit}...\n')

        values_src = Q_(data, src_unit)

        values_dst = values_src.to(dst_unit)  
            
        data_out[:] = values_dst

        return data_out




def square_two_top_cax(fx: float = 6.7, fy: float = 6.7, dpi: int = 300, 
                        layout: str = 'constrained', projection = None, frame: bool = False,
                        hspace: float = 0.3, wspace: float = 0.4, annotation: bool = True):

    import matplotlib.pyplot as plt
    from matplotlib.gridspec import GridSpec

    from my_.plot.style import style_1
    from my_.plot.init_ax import init_annotation_ax

    style_1()

    ncols, nrows = 2, 3

    figure = plt.figure(figsize = (fx, fy), 
                        dpi = dpi, )
                        #layout = layout)

    gs                              = GridSpec(figure = figure, 
                                               ncols = ncols, 
                                               nrows = nrows,
                                               height_ratios = [1, 20, 20],
                                               hspace = hspace, 
                                               wspace = wspace)

    axs_kw                          = {'frameon': frame, 'projection': projection}

    axs                             = gs.subplots(subplot_kw = axs_kw).flatten()

    if annotation: init_annotation_ax(axs[2:])

    return figure, axs[2:], axs[0:2]


def square_top_cax(fx: float = 6.7, fy: float = 6.7, dpi: int = 300, 
                        layout: str = 'constrained', projection = None, frame: bool = False,
                        hspace: float = 0.05, wspace: float = 0.05, annotation: bool = True,
                        x_an: float = 0.05, y_an: float = 1.05):

    import matplotlib.pyplot as plt
    from matplotlib.gridspec import GridSpec

    from my_.plot.style import style_1
    from my_.plot.init_ax import init_annotation_ax

    style_1()

    ncols, nrows                    = 2, 3

    figure                          = plt.figure(figsize = (fx, fy), dpi = dpi, layout = layout)

    gs                              = GridSpec(figure = figure, ncols = ncols, nrows = nrows,
                                               height_ratios = [1, 20, 20],
                                               hspace = hspace, wspace = wspace)

    axs_kw                          = {'frameon': frame, 'projection': projection}

    axs                             = gs.subplots(subplot_kw = axs_kw).flatten()

    for ax in axs[0:2]:
        ax.remove()

    if annotation: init_annotation_ax(axs[2:], x = x_an, y = y_an)
    
    axl                             = plt.subplot(gs[0,:], **axs_kw)

    return figure, axs[2:], axl





import numpy as np


def stack(array_list: list, 
          axis: int = 0) -> np.ndarray:

    """
    Stack arrays of the same shape
    Along a new axis (default 0)
    """

    stacked_array       = np.stack(array_list, axis = axis)

    return stacked_array


def select(array: np.ndarray, 
           indices: np.ndarray, 
           axis: int) -> np.ndarray:

    """
    Select values from array
    along axis dimension
    """

    new_shape           = [1] * array.ndim

    new_shape[axis]     = -1

    indices_dim         = np.reshape(indices, new_shape)

    array_selected      = np.take_along_axis(array, indices_dim, axis = axis)

    return array_selected


def select_by_other(array: np.ndarray, 
                    array_sel: np.ndarray, 
                    method: str, 
                    axis_select: int, 
                    axis_shrink: int) -> np.ndarray:

    """
    Select values from  a 3 or 4 dimenisonal array
    according to method applied on selector array
    The given axis dimension will be shrunk.
    Here assuming constant lat lon dimensions
    at axes -1, -2
    Currently 3d or 4d possible
    """

    from my_.grids.aggregate import apply

    indices             = apply(array_sel, axis_select, method)

    ndim                = array.ndim

    new_axes            = 0 if ndim == 3 else (0, 1) if ndim == 4 else None

    indices_4d          = np.expand_dims(indices, new_axes)

    array_selected      = np.take_along_axis(array, indices_4d, axis = axis_shrink)

    return array_selected


def select_by_self(array: np.ndarray, 
                   method: str, 
                   axis: int) -> np.ndarray:

    """
    Select values from array
    The given axis dimension will be shrunk.
    Here assuming constant lat lon dimensions
    at axes -1, -2
    Currently 3d or 4d possible
    """

    from my_.grids.aggregate import apply

    indices             = apply(array, axis, method)

    indices_dim         = np.expand_dims(indices, axis)

    array_selected      = np.take_along_axis(array, indices_dim, axis = axis)

    return array_selected


def sslice(array: np.ndarray, 
           indices: np.ndarray, 
           axis: int) -> np.ndarray:

    """
    Take the same slice for each 1d cell
    across axis dimension
    """

    array_sliced        = np.take(array, indices, axis = axis)

    return array_sliced


def select_apply_4d(array: np.ndarray, 
                    sel_func: callable = select_by_other, 
                    sel_func_args: dict = {}, 
                    apply_args: dict = {}) -> np.ndarray:

    """
    If there is a fourth dimension
    We select values along given axis
    By selector or by self
    And then we aggregate those by method sel
    and then we apply 
    """

    from my_.grids.aggregate import apply

    if array.ndim == 4:
         
        var_array_sel   = sel_func(array, **sel_func_args)

        array           = apply(var_array_sel, **apply_args)

    else: NotImplementedError

    return array


def grid_to_points(lat: np.ndarray, 
                   lon: np.ndarray) -> np.ndarray:

    stacked_coords              = np.dstack([lat, lon])
    
    list_points                 = stacked_coords.reshape(-1, 2)

    return list_points


def repeat_nd(array: np.ndarray,
              dims: int | list,
              repeats: dict,
              ndims: int =  2,) -> np.ndarray:
    
    if isinstance(dims, int): dims = [dims]

    ddims = [d for d in range(ndims) 
             if d not in dims]

    slx = [slice(0, None)] * ndims
    
    for d in ddims: slx[d] = None

    array = array[tuple(slx)]

    for d in ddims: 
        
        array = np.repeat(array, 
                          repeats[d], 
                          d)

    return array


import pandas as pd
from netCDF4 import Dataset

BGC_EUR0275_8daily = {'name': 'CLM5_DETECT',
                      'version': (0, 1, 0),
                      'path': '/p/data1/jibg31/CLM5_EUR-0275/BGC_outputs/carbon_control_DETECT/',
                      'type_file': 'netcdf',
                      'year_start': 1960,
                      'month_start': 1,
                      'year_end': 1989,
                      'month_end': 12,
                      'leapday': False,
                      'resolution_time': '8D',
                      'grid': 'EU3',
                      'variables': ['NEE', 
                                    'GPP', 
                                    'TLAI',
                                    'TOTECOSYSC',
                                    'TOTSOMC',
                                    'TOTVEGC',
                                    'LEAFC',
                                    'QFLX_EVAP_TOT',
                                    'QFLX_EVAP_VEG',
                                    'ZWT'],
                      'variable_names': {'NEE': 'NEE', 
                                         'GPP': 'GPP', 
                                         'TLAI': 'TLAI',
                                         'TOTECOSYSC': 'TOTECOSYSC',
                                         'TOTSOMC': 'TOTSOMC',
                                         'TOTVEGC': 'TOTVEGC',
                                         'LEAFC': 'LEAFC',
                                         'QFLX_EVAP_TOT': 'QFLX_EVAP_TOT',
                                         'QFLX_EVAP_VEG': 'QFLX_EVAP_VEG',
                                         'ZWT': 'ZWT'},
                      'variable_dimensions': {'NEE': ['time', 'lat', 'lon'], 
                                              'GPP': ['time', 'lat', 'lon'], 
                                              'TLAI': ['time', 'lat', 'lon'],
                                              'TOTECOSYSC': ['time', 'lat', 'lon'],
                                              'TOTSOMC': ['time', 'lat', 'lon'],
                                              'TOTVEGC': ['time', 'lat', 'lon'],
                                              'LEAFC': ['time', 'lat', 'lon'],
                                              'QFLX_EVAP_TOT': ['time', 'lat', 'lon'],
                                              'QFLX_EVAP_VEG': ['time', 'lat', 'lon'],
                                              'ZWT': ['time', 'lat', 'lon']},
                      'variable_units': {'NEE': 'g C / m^2 / s', 
                                         'GPP': 'g C / m^2 / s', 
                                         'TLAI': 'm^2 / m^2',
                                         'TOTECOSYSC': 'g C / m^2',
                                         'TOTSOMC': 'g C / m^2',
                                         'TOTVEGC': 'g C / m^2',
                                         'LEAFC': 'g C / m^2',
                                         'QFLX_EVAP_TOT': 'kg / m^2 / s',
                                         'QFLX_EVAP_VEG': 'mm H2O / s',
                                         'ZWT': 'm'},
                      'mask_value': None}

BGC_EUR0275_8daily_control = {'name': 'CLM5_DETECT_control',
                              'version': (0, 1, 0),
                              'path': '/p/data1/jibg31/CLM5_EUR-0275/BGC_outputs/carbon_control_DETECT/control/',
                              'type_file': 'netcdf',
                              'year_start': 1960,
                              'month_start': 1,
                              'year_end': 1989,
                              'month_end': 12,
                              'leapday': False,
                              'resolution_time': '8D',
                              'grid': 'EU3',
                              'variables': ['NEE', 
                                            'GPP', 
                                            'TLAI',
                                            'TOTECOSYSC',
                                            'TOTSOMC',
                                            'TOTVEGC',
                                            'LEAFC',
                                            'QFLX_EVAP_TOT',
                                            'QFLX_EVAP_VEG',
                                            'ZWT'],
                              'variable_names': {'NEE': 'NEE', 
                                                 'GPP': 'GPP', 
                                                 'TLAI': 'TLAI',
                                                 'TOTECOSYSC': 'TOTECOSYSC',
                                                 'TOTSOMC': 'TOTSOMC',
                                                 'TOTVEGC': 'TOTVEGC',
                                                 'LEAFC': 'LEAFC',
                                                 'QFLX_EVAP_TOT': 'QFLX_EVAP_TOT',
                                                 'QFLX_EVAP_VEG': 'QFLX_EVAP_VEG',
                                                 'ZWT': 'ZWT'},
                              'variable_dimensions': {'NEE': ['time', 'lat', 'lon'], 
                                                      'GPP': ['time', 'lat', 'lon'], 
                                                      'TLAI': ['time', 'lat', 'lon'],
                                                      'TOTECOSYSC': ['time', 'lat', 'lon'],
                                                      'TOTSOMC': ['time', 'lat', 'lon'],
                                                      'TOTVEGC': ['time', 'lat', 'lon'],
                                                      'LEAFC': ['time', 'lat', 'lon'],
                                                      'QFLX_EVAP_TOT': ['time', 'lat', 'lon'],
                                                      'QFLX_EVAP_VEG': ['time', 'lat', 'lon'],
                                                      'ZWT': ['time', 'lat', 'lon']},
                              'variable_units': {'NEE': 'g C / m^2 / s', 
                                                 'GPP': 'g C / m^2 / s', 
                                                 'TLAI': 'm^2 / m^2',
                                                 'TOTECOSYSC': 'g C / m^2',
                                                 'TOTSOMC': 'g C / m^2',
                                                 'TOTVEGC': 'g C / m^2',
                                                 'LEAFC': 'g C / m^2',
                                                 'QFLX_EVAP_TOT': 'kg / m^2 / s',
                                                 'QFLX_EVAP_VEG': 'mm H2O / s',
                                                 'ZWT': 'm'},
                              'mask_value': None}


BGC_EU3_8daily = {
    'name': 'BGC_EU3',
    'version': (1, 1, 7),
    'path': '/p/scratch/cjibg31/jibg3105/data/CLM5EU3/007/join_8d/',
    'type_file': 'netcdf',
    'year_start': 1995,
    'month_start': 1,
    'year_end': 2018,
    'month_end': 12,
    'resolution_time': '8D',
    'leapday': False,
    'grid': 'EU3',
    'variables': [
                    'GPP', 
                    'ET',
                    'ESOIL',
                    'Tr',
                    'BTRAN',
                    'Gs',
                    'Runoff',
                    'SM',
                    'WTD',
                    'H',
                    'SoiH',
                    'Rnet'
                 ],
    'variable_names': {
                        'GPP': 'GPP', 
                        'ET': 'QFLX_EVAP_TOT',
                        'ESOIL': 'QFLX_EVAP_GRND',
                        'Tr': 'QFLX_EVAP_VEG',
                        'BTRAN': 'BTRANMN',
                        'Gs': 'GSSHA',
                        'Runoff': 'QOVER',
                        'SM': 'H2OSOI',
                        'WTD': 'ZWT',
                        'H': 'Qh',
                        'SoiH': 'FGR',
                        'Rnet': 'Rnet',
                       },
    'variable_dimensions': {
                            'GPP': ['time', 'lat', 'lon'], 
                            'ET': ['time', 'lat', 'lon'],
                            'ESOIL': ['time', 'lat', 'lon'],
                            'Tr': ['time', 'lat', 'lon'],  
                            'BTRAN': ['time', 'lat', 'lon'],
                            'Gs': ['time', 'lat', 'lon'],
                            'Runoff': ['time', 'lat', 'lon'],
                            'SM': ['time', 'layer', 'lat', 'lon'],
                            'WTD': ['time', 'lat', 'lon'],
                            'H': ['time', 'lat', 'lon'],
                            'SoiH': ['time', 'lat', 'lon'],
                            'Rnet': ['time', 'lat', 'lon'],
                            }, 
    'variable_units': {'GPP': 'g/s',
                       'ET': 'mm/s',
                       'ESOIL': 'mm/s',
                       'Tr': 'mm/s',
                       'BTRAN': 'dimensionless',
                       'Gs': 'umol/s',
                       'Runoff': 'mm/s',
                       'SM': 'mm^3/mm^3',
                       'WTD': 'm',
                       'H': 'W/m^2',
                       'SoiH': 'W/m^2',
                       'Rnet': 'W/m^2'},
    'mask_value': None
}



BGC_EU3_daily = {
    'name': 'BGC_EU3',
    'version': (1, 1, 7),
    'path': '/p/data1/jibg31/CLM5.0_EU3km/outputs/007/join_d/',
    'type_file': 'netcdf',
    'year_start': 1995,
    'month_start': 1,
    'year_end': 2018,
    'month_end': 12,
    'resolution_time': 'D',
    'leapday': False,
    'grid': 'EU3',
    'variables': [
                    'GPP', 
                    'ET',
                    'ESOIL',
                    'Tr',
                    'BTRAN',
                    'Gs',
                    'Runoff',
                    'SM',
                    'WTD',
                    'H',
                    'SoiH',
                    'Rnet'
                 ],
    'variable_names': {
                        'GPP': 'GPP', 
                        'ET': 'QFLX_EVAP_TOT',
                        'ESOIL': 'QFLX_EVAP_GRND',
                        'Tr': 'QFLX_EVAP_VEG',
                        'BTRAN': 'BTRANMN',
                        'Gs': 'GSSHA',
                        'Runoff': 'QOVER',
                        'SM': 'H2OSOI',
                        'WTD': 'ZWT',
                        'H': 'Qh',
                        'SoiH': 'FGR',
                        'Rnet': 'Rnet',
                       },
    'variable_dimensions': {
                            'GPP': ['time', 'lat', 'lon'], 
                            'ET': ['time', 'lat', 'lon'],
                            'ESOIL': ['time', 'lat', 'lon'],
                            'Tr': ['time', 'lat', 'lon'],  
                            'BTRAN': ['time', 'lat', 'lon'],
                            'Gs': ['time', 'lat', 'lon'],
                            'Runoff': ['time', 'lat', 'lon'],
                            'SM': ['time', 'layer', 'lat', 'lon'],
                            'WTD': ['time', 'lat', 'lon'],
                            'H': ['time', 'lat', 'lon'],
                            'SoiH': ['time', 'lat', 'lon'],
                            'Rnet': ['time', 'lat', 'lon'],
                            }, 
    'variable_units': {'GPP': 'g/s',
                       'ET': 'mm/s',
                       'ESOIL': 'mm/s',
                       'Tr': 'mm/s',
                       'BTRAN': 'dimensionless',
                       'Gs': 'umol/s',
                       'Runoff': 'mm/s',
                       'SM': 'mm^3/mm^3',
                       'WTD': 'm',
                       'H': 'W/m^2',
                       'SoiH': 'W/m^2',
                       'Rnet': 'W/m^2'},
    'mask_value': None
}

BGC_EU3_weekly = {
    'name': 'BGC_EU3',
    'version': (1, 1, 7),
    'path': '/p/scratch/cjibg31/jibg3105/data/CLM5EU3/007/join_w/',
    'type_file': 'netcdf',
    'year_start': 1995,
    'month_start': 1,
    'year_end': 2018,
    'month_end': 12,
    'resolution_time': '7D',
    'leapday': False,
    'grid': 'EU3',
    'variables': [
                    'GPP', 
                    'ET',
                    'ESOIL',
                    'Tr',
                    'BTRAN',
                    'Gs',
                    'Runoff',
                    'SM',
                    'WTD',
                    'H',
                    'SoiH',
                    'Rnet'
                 ],
    'variable_names': {
                        'GPP': 'GPP', 
                        'ET': 'QFLX_EVAP_TOT',
                        'ESOIL': 'QFLX_EVAP_GRND',
                        'Tr': 'QFLX_EVAP_VEG',
                        'BTRAN': 'BTRANMN',
                        'Gs': 'GSSHA',
                        'Runoff': 'QOVER',
                        'SM': 'H2OSOI',
                        'WTD': 'ZWT',
                        'H': 'Qh',
                        'SoiH': 'FGR',
                        'Rnet': 'Rnet',
                       },
    'variable_dimensions': {
                            'GPP': ['time', 'lat', 'lon'], 
                            'ET': ['time', 'lat', 'lon'],
                            'ESOIL': ['time', 'lat', 'lon'],
                            'Tr': ['time', 'lat', 'lon'],  
                            'BTRAN': ['time', 'lat', 'lon'],
                            'Gs': ['time', 'lat', 'lon'],
                            'Runoff': ['time', 'lat', 'lon'],
                            'SM': ['time', 'layer', 'lat', 'lon'],
                            'WTD': ['time', 'lat', 'lon'],
                            'H': ['time', 'lat', 'lon'],
                            'SoiH': ['time', 'lat', 'lon'],
                            'Rnet': ['time', 'lat', 'lon'],
                            }, 
    'variable_units': {'GPP': 'g/s',
                       'ET': 'mm/s',
                       'ESOIL': 'mm/s',
                       'Tr': 'mm/s',
                       'BTRAN': 'dimensionless',
                       'Gs': 'umol/s',
                       'Runoff': 'mm/s',
                       'SM': 'mm^3/mm^3',
                       'WTD': 'm',
                       'H': 'W/m^2',
                       'SoiH': 'W/m^2',
                       'Rnet': 'W/m^2'},
    'mask_value': None
}


class CLM5_PFT:

    dir_surfdata: str = '/p/scratch/cjibg31/jibg3105/CESMDataRoot/InputData/lnd/clm2/surfdata_map/'
    file_surfdata: str = 'surfdata_CLM5EU3_v4_pos.nc'

    names = {1:'NET Temperate',
             2:'NET Boreal',
             3:'NDT Boreal',
             4:'BET Tropical',
             5:'BET Temperate',
             6:'BDT tropical',
             7:'BDT temperate',
             8:'BDT boreal',
             9:'BES temperate',
             10:'BDS temperate',
             11:'BDS boreal',
             12:'C3 grass arctic',
             13:'C3 grass',
             14:'C4 grass',
             15:'Crop R',
             16:'Crop I'}

    soil_layers = pd.DataFrame(data = [[0.0, 0.0, 0.0],
                                       [0.010,0.020,0.020],
                                       [0.040,0.040,0.060],
                                       [0.090,0.060,0.120],
                                       [0.160,0.080,0.200],
                                       [0.260,0.120,0.320],
                                       [0.400,0.160,0.480],
                                       [0.580,0.200,0.680],
                                       [0.800,0.240,0.920],
                                       [1.060,0.280,1.200],
                                       [1.360,0.320,1.520],
                                       [1.700,0.360,1.880],
                                       [2.080,0.400,2.280],
                                       [2.500,0.440,2.720],
                                       [2.990,0.540,3.260],
                                       [3.580,0.640,3.900],
                                       [4.270,0.740,4.640],
                                       [5.060,0.840,5.480],
                                       [5.950,0.940,6.420],
                                       [6.940,1.040,7.460],
                                       [8.030,1.140,8.600],
                                       [9.795,2.390,10.990],
                                       [13.328,4.676,15.666],
                                       [19.483,7.635,23.301],
                                       [28.871,11.140,34.441],
                                       [41.998,15.115,49.556]],
                               columns = ['node_depth',
                                          'thickness',
                                          'depth_interface'])
    
    parameter = {'betaroot': {1: 0.976,
                              2: 0.943,
                              3: 0.943,
                              4: 0.993,
                              5: 0.966,
                              6: 0.993,
                              7: 0.966,
                              8: 0.943,
                              9: 0.964,
                              10: 0.964,
                              11: 0.914,
                              12: 0.914,
                              13: 0.943,
                              14: 0.943,
                              15: 0.943,
                              16: 0.943}}

    def load_surface_var(self,
                         variables: str | list[str]):
        
        data = Dataset(f'{self.dir_surfdata}/{self.file_surfdata}')

        values = {v: data.variables[v][:] for v in variables}

        data.close()

        return values


import os
from datarie.handy import create_dirs, check_file_exists
from glob import glob
import xarray as xr

EUCORDEX_3km_8daily = {
    'name': 'COSMOREA5_EUCORDEX_3km_8daily',
    'version': (1, 0, 0),
    'path': '/p/scratch/cjibg31/jibg3105/data/COSMOREA6/8daily/',
    'type_file': 'netcdf',
    'year_start': 1995,
    'month_start': 1,
    'year_end': 2018,
    'month_end': 12,
    'resolution_time': '8D',
    'leapday': False,
    'grid': 'EU3',
    'variables': [
                  'P',
                  'Temp',
                  'PSRF',
                  'FSDS',
                  'FLDS',
                  'WIND', 
                  'RH',
                 ],
    'variable_names': {
                       'P': 'PRECTmms',
                       'Temp': 'TBOT',
                       'PSRF': 'PSRF',
                       'FSDS': 'FSDS',
                       'FLDS': 'FLDS',
                       'WIND': 'WIND',
                       'RH': 'RH',
                       },
    'variable_dimensions': {
                            'P': ['time', 'lat', 'lon'], 
                            'Temp': ['time', 'lat', 'lon'], 
                            'PSRF': ['time', 'lat', 'lon'], 
                            'FSDS': ['time', 'lat', 'lon'], 
                            'FLDS': ['time', 'lat', 'lon'], 
                            'WIND': ['time', 'lat', 'lon'], 
                            'RH': ['time', 'lat', 'lon'], 
                           }, 
    'variable_units': {
                       'P': 'mm/s',
                       'Temp': 'K',
                       'PSRF': 'Pa',
                       'FSDS': 'W/m^2',
                       'FLDS': 'W/m^2',
                       'WIND': 'm/s',
                       'RH': '%',
                      },
    'mask_value': None
}

EUCORDEX_3km_weekly = {
    'name': 'COSMOREA5_EUCORDEX_3km_weekly',
    'version': (1, 0, 0),
    'path': '/p/scratch/cjibg31/jibg3105/data/COSMOREA6/weekly/',
    'type_file': 'netcdf',
    'year_start': 1995,
    'month_start': 1,
    'year_end': 2018,
    'month_end': 12,
    'resolution_time': '7D',
    'leapday': False,
    'grid': 'EU3',
    'variables': [
                  'P',
                  'Temp',
                  'PSRF',
                  'FSDS',
                  'FLDS',
                  'WIND', 
                  'RH',
                 ],
    'variable_names': {
                       'P': 'PRECTmms',
                       'Temp': 'TBOT',
                       'PSRF': 'PSRF',
                       'FSDS': 'FSDS',
                       'FLDS': 'FLDS',
                       'WIND': 'WIND',
                       'RH': 'RH',
                       },
    'variable_dimensions': {
                            'P': ['time', 'lat', 'lon'], 
                            'Temp': ['time', 'lat', 'lon'], 
                            'PSRF': ['time', 'lat', 'lon'], 
                            'FSDS': ['time', 'lat', 'lon'], 
                            'FLDS': ['time', 'lat', 'lon'], 
                            'WIND': ['time', 'lat', 'lon'], 
                            'RH': ['time', 'lat', 'lon'], 
                           }, 
    'variable_units': {
                       'P': 'mm/s',
                       'Temp': 'K',
                       'PSRF': 'Pa',
                       'FSDS': 'W/m^2',
                       'FLDS': 'W/m^2',
                       'WIND': 'm/s',
                       'RH': '%',
                      },
    'mask_value': None
}


EUCORDEX_3km_6hourly = {
    'name': 'COSMOREA5_EUCORDEX_3km_6hourly',
    'version': (1, 0, 0),
    'path': '/p/data1/jibg31/FORCINGS/COSMOREA6/yearly_files/',
    'type_file': 'netcdf',
    'year_start': 1995,
    'month_start': 1,
    'year_end': 2018,
    'month_end': 12,
    'resolution_time': '6H',
    'leapday': False,
    'grid': 'EU3',
    'variables': [
                  'P',
                  'Temp',
                  'PSRF',
                  'FSDS',
                  'FLDS',
                  'WIND', 
                  'RH',
                 ],
    'variable_names': {
                       'P': 'PRECTmms',
                       'Temp': 'TBOT',
                       'PSRF': 'PSRF',
                       'FSDS': 'FSDS',
                       'FLDS': 'FLDS',
                       'WIND': 'WIND',
                       'RH': 'RH',
                       },
    'variable_dimensions': {
                            'P': ['time', 'lat', 'lon'], 
                            'Temp': ['time', 'lat', 'lon'], 
                            'PSRF': ['time', 'lat', 'lon'], 
                            'FSDS': ['time', 'lat', 'lon'], 
                            'FLDS': ['time', 'lat', 'lon'], 
                            'WIND': ['time', 'lat', 'lon'], 
                            'RH': ['time', 'lat', 'lon'], 
                           }, 
    'variable_units': {
                       'P': 'mm/s',
                       'Temp': 'K',
                       'PSRF': 'Pa',
                       'FSDS': 'W/m^2',
                       'FLDS': 'W/m^2',
                       'WIND': 'm/s',
                       'RH': '%',
                      },
    'mask_value': None
}


def create_yearly_files(path_rawdata: str,
                        path_out: str):
    
    """
    Input directory should only contain the ERA5L rawdata files...
    """
    
    create_dirs(path_out)

    files = sorted(glob(f'{path_rawdata}/*.nc'))

    for f in files:

        file_year = f.split('/')[-1]

        if check_file_exists(f'{path_out}/{file_year}'): continue

        data_raw = xr.open_dataset(f)

        print(f'Create yearly file for file {f}...\n')

        data_raw.to_netcdf(f'{path_out}/{file_year}',
                         format = 'NETCDF4_CLASSIC', 
                         unlimited_dims = ['time'])


import pandas as pd
    
def multi_columns(groups: list = [], names: list = []):

    import pandas as pd

    multi_columns           = pd.MultiIndex.from_product(groups, names = names)

    return multi_columns


def src_var_lc_station_index(sources, variables, landcover, stations):

    import pandas as pd

    columns                 = multi_columns(groups = [sources, variables, stations], names = ['Source', 'Variable', 'Station'])

    df_columns              = columns.to_frame()

    df_columns.insert(2, 'Landcover', landcover)

    new_columns             = pd.MultiIndex.from_frame(df_columns)

    new_columns_sorted      = new_columns.sort_values()

    return new_columns_sorted


def single_src_var_X_index(array, src, var, lc, name_station):

    if array.ndim == 3: 

        column          = single_src_var_lc_station_index(src, var, lc, name_station)

    elif array.ndim == 4:

        n_layers                = array.shape[1]

        column          = single_src_var_lc_station_layer_index(src, var, lc, name_station, n_layers = n_layers)

    return column


def single_src_var_lc_station_index(src, var, lc, station):
    
    import pandas as pd

    names                   = ['Source', 'Variable', 'Landcover', 'Station']

    tuple_index             = (src, var, lc, station)

    index                   = pd.MultiIndex.from_tuples([tuple_index], names = names)

    return index


def single_multiindex(dictionary: dict):
    
    import pandas as pd

    names = list(dictionary.keys())

    indices = list(dictionary.values())

    if any([isinstance(i, list) for i in indices]):

        indices = [[i] if not isinstance(i, list)
                   else i
                   for i in indices]

        index = pd.MultiIndex.from_product(indices, names = names)

    else:
        
        index = pd.MultiIndex.from_tuples([indices], names = names)

    return index


def single_src_var_lc_station_layer_index(src, var, lc, station, n_layers):
    
    import pandas as pd
    import numpy as np

    if n_layers > 0:

        layers              = np.arange(n_layers).astype(str).tolist()

    elif n_layers < 0:

        layers              = [n_layers]

    names                   = ['Source', 'Variable', 'Landcover', 'Station', 'Layer']

    tuple_index             = [[src], [var], [lc], [station], layers]

    index                   = pd.MultiIndex.from_product(tuple_index, names = names)

    return index


def add_level_to_multi_columns(df, level: int, name_level: str, level_values, sort = False):

    import pandas as pd

    # Convert index to dataframe
    old_idx             = df.columns.to_frame()

    # Insert new level at specified location

    if level == -1:

        new_level       = pd.Series(level_values, name = name_level, index = old_idx.index)

        old_idx         = old_idx.join(new_level)

    else: 
        
        old_idx.insert(level, name_level, level_values)

    # Convert back to MultiIndex
    new_index           = pd.MultiIndex.from_frame(old_idx)

    df_new              = df.copy()
    df_new.columns      = new_index

    if sort: df_new == df_new.sort_index(axis = 1)

    return df_new


def layer_level_to_int(df):

    level_names         = df.columns.names

    if 'Layer' not in level_names: return df
    
    i_level             = level_names.index('Layer')

    level_int           = df.columns.levels[i_level].astype(int)
    
    df.columns          = df.columns.set_levels(level_int, level = i_level)

    return df


def select_multi_index(df: pd.DataFrame, 
                       levels: list, 
                       keys: list, 
                       axis = 1) -> pd.DataFrame:

    if isinstance(levels, str): levels = [levels]
    if isinstance(keys, str): keys = [keys]
    if isinstance(df, pd.Series): df = df.to_frame()

    axx = df.columns if axis == 1 else df.index

    n_levels = axx.nlevels

    names_levels = axx.names

    index_level = [names_levels.index(level) for level in levels]

    list_index = [slice(None)] * n_levels

    for i_indxs, indxs in enumerate(index_level):
        
        kk = keys[i_indxs]

        if isinstance(kk, str): kk = [kk]

        iikeys = [k for k in kk
                  if k in axx.get_level_values(indxs)]

        list_index[indxs] = iikeys
    
    if axis == 1:
        df_sel = df.loc[:, tuple(list_index)]

    else: 
        df_sel = df.loc[tuple(list_index), :]

    return df_sel


def nona_level(df: pd.DataFrame, 
               level: str | list[str], 
               axis: int = 1, 
               how: str = 'all'):

    if axis == 1 : df = df.T

    df_level_groups = df.groupby(sort = False, 
                                 level = level)
    
    df_nona = df_level_groups.apply(pd.DataFrame.dropna,
                                    axis = 1,
                                    how = how)
    
    if axis == 1: return df_nona.T

    return df_nona 





sec_per_day                     = 60*60*24


def transform(values, source: str, variable: str):

    """
    Process an array linearly
    with scalar multiplicator and addend
    given by variable parameter dictionary
    """

    from my_.resources.sources import query_unit_transform
    
    ux_method                   = query_unit_transform(source, variable, 'methods')

    if ux_method == 'linear':

        from my_.math.mapping import linear

        m                       = query_unit_transform(source, variable, 'm')
        b                       = query_unit_transform(source, variable, 'b')

        values_uxb              = linear(values, m, b)

    return values_uxb


def transform_df(df, source: str, variables: list):
    
    from my_.resources.sources import query_unit_transform

    df_out                      = df.copy()
    
    for i_col, col in enumerate(df.columns):

        variable                = variables[i_col]

        ux_method               = query_unit_transform(source, variable, 'methods')

        if ux_method == 'linear':

            from my_.math.mapping import linear

            m                       = query_unit_transform(source, variable, 'm')
            b                       = query_unit_transform(source, variable, 'b')

            df_out[col]              = linear(df[col].astype('float'), m, b)

    return df_out


from dataclasses import dataclass
import pandas as pd
from glob import glob
import xarray as xr
import numpy as np
import os 
from typing import Any

from datarie.handy import create_dirs, check_file_exists
from datarie.templates import gridded_data

@dataclass(kw_only = True)
class cluster_data(gridded_data):
    
    agg_period: str
    filter_area: float # [km²]
    weigh_mid: float
    weigh_steep: float
    grid_window: int
    frac_points: float
    factor_time: float
    km_per_cell: float = 9


    def __post_init__(self):
        
        self.path = self.path + \
                    f'/{self.agg_period}_' + \
                    f'l-{self.grid_window}_' + \
                    f'p-{self.frac_points}_' + \
                    f'k-{self.weigh_mid}_' + \
                    f'e-{self.weigh_steep}_' + \
                    f'a-{self.filter_area}_' + \
                    f't-{self.factor_time}_' + \
                    f'y0-{self.year_start}_' + \
                    f'y1-{self.year_end}/'

    def get_file_fix(self):
        
        str_f = f'{self.agg_period}_' + \
                f'l-{self.grid_window}_' + \
                f'p-{self.frac_points}_' + \
                f'k-{self.weigh_mid}_' + \
                f'e-{self.weigh_steep}_' + \
                f'a-{self.filter_area}_' + \
                f't-{self.factor_time}_' + \
                f'y0-{self.year_start}_' + \
                f'y1-{self.year_end}'
        
        return str_f
    
    def get_properties(self, 
                       variable: str):

        dir_file = '/p/scratch/cjibg31/jibg3105/projects/papers/CLM5EU3_Droughts/out/cluster_properties'

        name_file = self.get_file_fix()

        properties = pd.read_excel(f'{dir_file}/{variable}_{name_file}.xlsx', 
                           header = [0, 1],
                           index_col = 0)
        
        return properties
    
variables_EU3_8daily = {
    'name': 'HOLIDROUGHT_BGC_EU3_variables',
    'version': (0, 1, 0),
    'path': '/p/data1/jibg31/poppe1/HOLIDROUGHT/variables/',
    'type_file': 'netcdf',
    'year_start': 1995,
    'month_start': 1,
    'year_end': 2018,
    'month_end': 12,
    'leapday': False,
    'resolution_time': '8D',
    'grid': 'EU3',
    'variables': ['GPP', 
                  'ESOIL',
                  'Tr',
                  'BTRAN',
                  'Gs',
                  'Runoff',
                  'SM',
                  'WTD',
                  'VPD',
                  'P'],
    'variable_names': {'GPP': 'GPP', 
                       'ESOIL': 'ESOIL',
                       'Tr': 'Tr',
                       'BTRAN': 'BTRAN',
                       'Gs': 'Gs',
                       'Runoff': 'Runoff',
                       'SM': 'SM',
                       'WTD': 'WTD',
                       'VPD': 'VPD',
                       'P': 'P'},
    'variable_dimensions': {'GPP': ['time', 'lat', 'lon'], 
                            'ESOIL': ['time', 'lat', 'lon'],
                            'Tr': ['time', 'lat', 'lon'],  
                            'BTRAN': ['time', 'lat', 'lon'],
                            'Gs': ['time', 'lat', 'lon'],
                            'Runoff': ['time', 'lat', 'lon'],
                            'SM': ['time', 'layer', 'lat', 'lon'],
                            'WTD': ['time', 'lat', 'lon'],
                            'VPD': ['time', 'lat', 'lon'],
                            'P': ['time', 'lat', 'lon']}, 
    'variable_units': {'GPP': 'g/day',
                       'ESOIL': 'mm/day',
                       'Tr': 'mm/day',
                       'BTRAN': 'dimensionless',
                       'Gs': 'umol/day',
                       'Runoff': 'mm/day',
                       'SM': 'm^3/m^3',
                       'WTD': 'm',
                       'VPD': 'kPa',
                       'P': 'mm/day'},
    'mask_value': None
}


SXI_365D_EU3_8daily = {
    'name': 'HOLIDROUGHT_SXI_365D',
    'version': (0, 1, 0),
    'path': '/p/data1/jibg31/poppe1/HOLIDROUGHT/SXI/365D/',
    'type_file': 'netcdf',
    'year_start': 1996,
    'month_start': 1,
    'year_end': 2018,
    'month_end': 12,
    'resolution_time': '8D',
    'grid': 'EU3',
    'leapday': False,
    'variables': ['SXI_GPP', 
                  'SXI_ESOIL',
                  'SXI_Tr',
                  'SXI_BTRAN',
                  'SXI_Gs',
                  'SXI_Runoff',
                  'SXI_SM_0',
                  'SXI_SM_1',
                  'SXI_SM_2',
                  'SXI_SM_3',
                  'SXI_SM_4',
                  'SXI_SM_5',
                  'SXI_SM_6',
                  'SXI_SM_7',
                  'SXI_SM_8',
                  'SXI_SM_9',
                  'SXI_SM_10',
                  'SXI_SM_11',
                  'SXI_SM_12',
                  'SXI_SM_13',
                  'SXI_SM_14',
                  'SXI_SM_15',
                  'SXI_SM_16',
                  'SXI_SM_17',
                  'SXI_SM_18',
                  'SXI_SM_19',
                  'SXI_WTD',
                  'SXI_VPD',
                  'SXI_P',
                  'SXI_SM_r'],
    'variable_names': {'SXI_GPP': 'SXI_GPP', 
                       'SXI_ESOIL': 'SXI_ESOIL',
                       'SXI_Tr': 'SXI_Tr',
                       'SXI_BTRAN': 'SXI_BTRAN',
                       'SXI_Gs': 'SXI_Gs',
                       'SXI_Runoff': 'SXI_Runoff',
                       'SXI_SM_0': 'SXI_SM_0',
                       'SXI_SM_1': 'SXI_SM_1',
                       'SXI_SM_2': 'SXI_SM_2',
                       'SXI_SM_3': 'SXI_SM_3',
                       'SXI_SM_4': 'SXI_SM_4',
                       'SXI_SM_5': 'SXI_SM_5',
                       'SXI_SM_6': 'SXI_SM_6',
                       'SXI_SM_7': 'SXI_SM_7',
                       'SXI_SM_8': 'SXI_SM_8',
                       'SXI_SM_9': 'SXI_SM_9',
                       'SXI_SM_10': 'SXI_SM_10',
                       'SXI_SM_11': 'SXI_SM_11',
                       'SXI_SM_12': 'SXI_SM_12',
                       'SXI_SM_13': 'SXI_SM_13',
                       'SXI_SM_14': 'SXI_SM_14',
                       'SXI_SM_15': 'SXI_SM_15',
                       'SXI_SM_16': 'SXI_SM_16',
                       'SXI_SM_17': 'SXI_SM_17',
                       'SXI_SM_18': 'SXI_SM_18',
                       'SXI_SM_19': 'SXI_SM_19',
                       'SXI_WTD': 'SXI_WTD',
                       'SXI_VPD': 'SXI_VPD',
                       'SXI_P': 'SXI_P',
                       'SXI_SM_r': 'SXI_SM_r'},
    'variable_dimensions': {'SXI_GPP': ['time', 'lat', 'lon'], 
                            'SXI_ESOIL': ['time', 'lat', 'lon'],
                            'SXI_Tr': ['time', 'lat', 'lon'],  
                            'SXI_BTRAN': ['time', 'lat', 'lon'],
                            'SXI_Gs': ['time', 'lat', 'lon'],
                            'SXI_Runoff': ['time', 'lat', 'lon'],
                            'SXI_SM_0': ['time', 'lat', 'lon'],
                            'SXI_SM_1': ['time', 'lat', 'lon'],
                            'SXI_SM_2': ['time', 'lat', 'lon'],
                            'SXI_SM_3': ['time', 'lat', 'lon'],
                            'SXI_SM_4': ['time', 'lat', 'lon'],
                            'SXI_SM_5': ['time', 'lat', 'lon'],
                            'SXI_SM_6': ['time', 'lat', 'lon'],
                            'SXI_SM_7': ['time', 'lat', 'lon'],
                            'SXI_SM_8': ['time', 'lat', 'lon'],
                            'SXI_SM_9': ['time', 'lat', 'lon'],
                            'SXI_SM_10': ['time', 'lat', 'lon'],
                            'SXI_SM_11': ['time', 'lat', 'lon'],
                            'SXI_SM_12': ['time', 'lat', 'lon'],
                            'SXI_SM_13': ['time', 'lat', 'lon'],
                            'SXI_SM_14': ['time', 'lat', 'lon'],
                            'SXI_SM_15': ['time', 'lat', 'lon'],
                            'SXI_SM_16': ['time', 'lat', 'lon'],
                            'SXI_SM_17': ['time', 'lat', 'lon'],
                            'SXI_SM_18': ['time', 'lat', 'lon'],
                            'SXI_SM_19': ['time', 'lat', 'lon'],
                            'SXI_WTD': ['time', 'lat', 'lon'],
                            'SXI_VPD': ['time', 'lat', 'lon'],
                            'SXI_P': ['time', 'lat', 'lon'],
                            'SXI_SM_r': ['time', 'lat', 'lon'],}, 
    'variable_units': {'SXI_GPP': 'dimensionless',
                       'SXI_ESOIL': 'dimensionless',
                       'SXI_Tr': 'dimensionless',
                       'SXI_BTRAN': 'dimensionless',
                       'SXI_Gs': 'dimensionless',
                       'SXI_Runoff': 'dimensionless',
                       'SXI_SM_0': 'dimensionless',
                       'SXI_SM_1': 'dimensionless',
                       'SXI_SM_2': 'dimensionless',
                       'SXI_SM_3': 'dimensionless',
                       'SXI_SM_4': 'dimensionless',
                       'SXI_SM_5': 'dimensionless',
                       'SXI_SM_6': 'dimensionless',
                       'SXI_SM_7': 'dimensionless',
                       'SXI_SM_8': 'dimensionless',
                       'SXI_SM_9': 'dimensionless',
                       'SXI_SM_10': 'dimensionless',
                       'SXI_SM_11': 'dimensionless',
                       'SXI_SM_12': 'dimensionless',
                       'SXI_SM_13': 'dimensionless',
                       'SXI_SM_14': 'dimensionless',
                       'SXI_SM_15': 'dimensionless',
                       'SXI_SM_16': 'dimensionless',
                       'SXI_SM_17': 'dimensionless',
                       'SXI_SM_18': 'dimensionless',
                       'SXI_SM_19': 'dimensionless',
                       'SXI_WTD': 'dimensionless',
                       'SXI_VPD': 'dimensionless',
                       'SXI_P': 'dimensionless',
                       'SXI_SM_r': 'dimensionless'},
    'mask_value': None
}


SXI_183D_EU3_8daily = {
    'name': 'HOLIDROUGHT_SXI_183D',
    'version': (0, 1, 0),
    'path': '/p/data1/jibg31/poppe1/HOLIDROUGHT/SXI/183D/',
    'type_file': 'netcdf',
    'year_start': 1995,
    'month_start': 7,
    'year_end': 2018,
    'month_end': 12,
    'leapday': False,
    'resolution_time': '8D',
    'grid': 'EU3',
    'variables': ['SXI_GPP', 
                  'SXI_ESOIL',
                  'SXI_Tr',
                  'SXI_BTRAN',
                  'SXI_Gs',
                  'SXI_Runoff',
                  'SXI_SM_0',
                  'SXI_SM_1',
                  'SXI_SM_2',
                  'SXI_SM_3',
                  'SXI_SM_4',
                  'SXI_SM_5',
                  'SXI_SM_6',
                  'SXI_SM_7',
                  'SXI_SM_8',
                  'SXI_SM_9',
                  'SXI_SM_10',
                  'SXI_SM_11',
                  'SXI_SM_12',
                  'SXI_SM_13',
                  'SXI_SM_14',
                  'SXI_SM_15',
                  'SXI_SM_16',
                  'SXI_SM_17',
                  'SXI_SM_18',
                  'SXI_SM_19',
                  'SXI_WTD',
                  'SXI_VPD',
                  'SXI_P',
                  'SXI_SM_r'],
    'variable_names': {'SXI_GPP': 'SXI_GPP', 
                       'SXI_ESOIL': 'SXI_ESOIL',
                       'SXI_Tr': 'SXI_Tr',
                       'SXI_BTRAN': 'SXI_BTRAN',
                       'SXI_Gs': 'SXI_Gs',
                       'SXI_Runoff': 'SXI_Runoff',
                       'SXI_SM_0': 'SXI_SM_0',
                       'SXI_SM_1': 'SXI_SM_1',
                       'SXI_SM_2': 'SXI_SM_2',
                       'SXI_SM_3': 'SXI_SM_3',
                       'SXI_SM_4': 'SXI_SM_4',
                       'SXI_SM_5': 'SXI_SM_5',
                       'SXI_SM_6': 'SXI_SM_6',
                       'SXI_SM_7': 'SXI_SM_7',
                       'SXI_SM_8': 'SXI_SM_8',
                       'SXI_SM_9': 'SXI_SM_9',
                       'SXI_SM_10': 'SXI_SM_10',
                       'SXI_SM_11': 'SXI_SM_11',
                       'SXI_SM_12': 'SXI_SM_12',
                       'SXI_SM_13': 'SXI_SM_13',
                       'SXI_SM_14': 'SXI_SM_14',
                       'SXI_SM_15': 'SXI_SM_15',
                       'SXI_SM_16': 'SXI_SM_16',
                       'SXI_SM_17': 'SXI_SM_17',
                       'SXI_SM_18': 'SXI_SM_18',
                       'SXI_SM_19': 'SXI_SM_19',
                       'SXI_WTD': 'SXI_WTD',
                       'SXI_VPD': 'SXI_VPD',
                       'SXI_P': 'SXI_P',
                       'SXI_SM_r': 'SXI_SM_r'},
    'variable_dimensions': {'SXI_GPP': ['time', 'lat', 'lon'], 
                            'SXI_ESOIL': ['time', 'lat', 'lon'],
                            'SXI_Tr': ['time', 'lat', 'lon'],  
                            'SXI_BTRAN': ['time', 'lat', 'lon'],
                            'SXI_Gs': ['time', 'lat', 'lon'],
                            'SXI_Runoff': ['time', 'lat', 'lon'],
                            'SXI_SM_0': ['time', 'lat', 'lon'],
                            'SXI_SM_1': ['time', 'lat', 'lon'],
                            'SXI_SM_2': ['time', 'lat', 'lon'],
                            'SXI_SM_3': ['time', 'lat', 'lon'],
                            'SXI_SM_4': ['time', 'lat', 'lon'],
                            'SXI_SM_5': ['time', 'lat', 'lon'],
                            'SXI_SM_6': ['time', 'lat', 'lon'],
                            'SXI_SM_7': ['time', 'lat', 'lon'],
                            'SXI_SM_8': ['time', 'lat', 'lon'],
                            'SXI_SM_9': ['time', 'lat', 'lon'],
                            'SXI_SM_10': ['time', 'lat', 'lon'],
                            'SXI_SM_11': ['time', 'lat', 'lon'],
                            'SXI_SM_12': ['time', 'lat', 'lon'],
                            'SXI_SM_13': ['time', 'lat', 'lon'],
                            'SXI_SM_14': ['time', 'lat', 'lon'],
                            'SXI_SM_15': ['time', 'lat', 'lon'],
                            'SXI_SM_16': ['time', 'lat', 'lon'],
                            'SXI_SM_17': ['time', 'lat', 'lon'],
                            'SXI_SM_18': ['time', 'lat', 'lon'],
                            'SXI_SM_19': ['time', 'lat', 'lon'],
                            'SXI_WTD': ['time', 'lat', 'lon'],
                            'SXI_VPD': ['time', 'lat', 'lon'],
                            'SXI_P': ['time', 'lat', 'lon'],
                            'SXI_SM_r': ['time', 'lat', 'lon'],}, 
    'variable_units': {'SXI_GPP': 'dimensionless',
                       'SXI_ESOIL': 'dimensionless',
                       'SXI_Tr': 'dimensionless',
                       'SXI_BTRAN': 'dimensionless',
                       'SXI_Gs': 'dimensionless',
                       'SXI_Runoff': 'dimensionless',
                       'SXI_SM_0': 'dimensionless',
                       'SXI_SM_1': 'dimensionless',
                       'SXI_SM_2': 'dimensionless',
                       'SXI_SM_3': 'dimensionless',
                       'SXI_SM_4': 'dimensionless',
                       'SXI_SM_5': 'dimensionless',
                       'SXI_SM_6': 'dimensionless',
                       'SXI_SM_7': 'dimensionless',
                       'SXI_SM_8': 'dimensionless',
                       'SXI_SM_9': 'dimensionless',
                       'SXI_SM_10': 'dimensionless',
                       'SXI_SM_11': 'dimensionless',
                       'SXI_SM_12': 'dimensionless',
                       'SXI_SM_13': 'dimensionless',
                       'SXI_SM_14': 'dimensionless',
                       'SXI_SM_15': 'dimensionless',
                       'SXI_SM_16': 'dimensionless',
                       'SXI_SM_17': 'dimensionless',
                       'SXI_SM_18': 'dimensionless',
                       'SXI_SM_19': 'dimensionless',
                       'SXI_WTD': 'dimensionless',
                       'SXI_VPD': 'dimensionless',
                       'SXI_P': 'dimensionless',
                       'SXI_SM_r': 'dimensionless'},
    'mask_value': None
}

SXI_92D_EU3_8daily: dict[str, Any] = {
    'name': 'HOLIDROUGHT_SXI_92D',
    'version': (0, 1, 0),
    'path': '/p/data1/jibg31/poppe1/HOLIDROUGHT/SXI/92D/',
    'type_file': 'netcdf',
    'year_start': 1995,
    'month_start': 4,
    'year_end': 2018,
    'month_end': 12,
    'leapday': False,
    'resolution_time': '8D',
    'grid': 'EU3',
    'variables': ['SXI_GPP', 
                  'SXI_ESOIL',
                  'SXI_Tr',
                  'SXI_BTRAN',
                  'SXI_Gs',
                  'SXI_Runoff',
                  'SXI_SM_0',
                  'SXI_SM_1',
                  'SXI_SM_2',
                  'SXI_SM_3',
                  'SXI_SM_4',
                  'SXI_SM_5',
                  'SXI_SM_6',
                  'SXI_SM_7',
                  'SXI_SM_8',
                  'SXI_SM_9',
                  'SXI_SM_10',
                  'SXI_SM_11',
                  'SXI_SM_12',
                  'SXI_SM_13',
                  'SXI_SM_14',
                  'SXI_SM_15',
                  'SXI_SM_16',
                  'SXI_SM_17',
                  'SXI_SM_18',
                  'SXI_SM_19',
                  'SXI_WTD',
                  'SXI_VPD',
                  'SXI_P',
                  'SXI_SM_r'],
    'variable_names': {'SXI_GPP': 'SXI_GPP', 
                       'SXI_ESOIL': 'SXI_ESOIL',
                       'SXI_Tr': 'SXI_Tr',
                       'SXI_BTRAN': 'SXI_BTRAN',
                       'SXI_Gs': 'SXI_Gs',
                       'SXI_Runoff': 'SXI_Runoff',
                       'SXI_SM_0': 'SXI_SM_0',
                       'SXI_SM_1': 'SXI_SM_1',
                       'SXI_SM_2': 'SXI_SM_2',
                       'SXI_SM_3': 'SXI_SM_3',
                       'SXI_SM_4': 'SXI_SM_4',
                       'SXI_SM_5': 'SXI_SM_5',
                       'SXI_SM_6': 'SXI_SM_6',
                       'SXI_SM_7': 'SXI_SM_7',
                       'SXI_SM_8': 'SXI_SM_8',
                       'SXI_SM_9': 'SXI_SM_9',
                       'SXI_SM_10': 'SXI_SM_10',
                       'SXI_SM_11': 'SXI_SM_11',
                       'SXI_SM_12': 'SXI_SM_12',
                       'SXI_SM_13': 'SXI_SM_13',
                       'SXI_SM_14': 'SXI_SM_14',
                       'SXI_SM_15': 'SXI_SM_15',
                       'SXI_SM_16': 'SXI_SM_16',
                       'SXI_SM_17': 'SXI_SM_17',
                       'SXI_SM_18': 'SXI_SM_18',
                       'SXI_SM_19': 'SXI_SM_19',
                       'SXI_WTD': 'SXI_WTD',
                       'SXI_VPD': 'SXI_VPD',
                       'SXI_P': 'SXI_P',
                       'SXI_SM_r': 'SXI_SM_r'},
    'variable_dimensions': {'SXI_GPP': ['time', 'lat', 'lon'], 
                            'SXI_ESOIL': ['time', 'lat', 'lon'],
                            'SXI_Tr': ['time', 'lat', 'lon'],  
                            'SXI_BTRAN': ['time', 'lat', 'lon'],
                            'SXI_Gs': ['time', 'lat', 'lon'],
                            'SXI_Runoff': ['time', 'lat', 'lon'],
                            'SXI_SM_0': ['time', 'lat', 'lon'],
                            'SXI_SM_1': ['time', 'lat', 'lon'],
                            'SXI_SM_2': ['time', 'lat', 'lon'],
                            'SXI_SM_3': ['time', 'lat', 'lon'],
                            'SXI_SM_4': ['time', 'lat', 'lon'],
                            'SXI_SM_5': ['time', 'lat', 'lon'],
                            'SXI_SM_6': ['time', 'lat', 'lon'],
                            'SXI_SM_7': ['time', 'lat', 'lon'],
                            'SXI_SM_8': ['time', 'lat', 'lon'],
                            'SXI_SM_9': ['time', 'lat', 'lon'],
                            'SXI_SM_10': ['time', 'lat', 'lon'],
                            'SXI_SM_11': ['time', 'lat', 'lon'],
                            'SXI_SM_12': ['time', 'lat', 'lon'],
                            'SXI_SM_13': ['time', 'lat', 'lon'],
                            'SXI_SM_14': ['time', 'lat', 'lon'],
                            'SXI_SM_15': ['time', 'lat', 'lon'],
                            'SXI_SM_16': ['time', 'lat', 'lon'],
                            'SXI_SM_17': ['time', 'lat', 'lon'],
                            'SXI_SM_18': ['time', 'lat', 'lon'],
                            'SXI_SM_19': ['time', 'lat', 'lon'],
                            'SXI_WTD': ['time', 'lat', 'lon'],
                            'SXI_VPD': ['time', 'lat', 'lon'],
                            'SXI_P': ['time', 'lat', 'lon'],
                            'SXI_SM_r': ['time', 'lat', 'lon'],}, 
    'variable_units': {'SXI_GPP': 'dimensionless',
                       'SXI_ESOIL': 'dimensionless',
                       'SXI_Tr': 'dimensionless',
                       'SXI_BTRAN': 'dimensionless',
                       'SXI_Gs': 'dimensionless',
                       'SXI_Runoff': 'dimensionless',
                       'SXI_SM_0': 'dimensionless',
                       'SXI_SM_1': 'dimensionless',
                       'SXI_SM_2': 'dimensionless',
                       'SXI_SM_3': 'dimensionless',
                       'SXI_SM_4': 'dimensionless',
                       'SXI_SM_5': 'dimensionless',
                       'SXI_SM_6': 'dimensionless',
                       'SXI_SM_7': 'dimensionless',
                       'SXI_SM_8': 'dimensionless',
                       'SXI_SM_9': 'dimensionless',
                       'SXI_SM_10': 'dimensionless',
                       'SXI_SM_11': 'dimensionless',
                       'SXI_SM_12': 'dimensionless',
                       'SXI_SM_13': 'dimensionless',
                       'SXI_SM_14': 'dimensionless',
                       'SXI_SM_15': 'dimensionless',
                       'SXI_SM_16': 'dimensionless',
                       'SXI_SM_17': 'dimensionless',
                       'SXI_SM_18': 'dimensionless',
                       'SXI_SM_19': 'dimensionless',
                       'SXI_WTD': 'dimensionless',
                       'SXI_VPD': 'dimensionless',
                       'SXI_P': 'dimensionless',
                       'SXI_SM_r': 'dimensionless'},
    'mask_value': None
}


SXI_31D_EU3_8daily = {
    'name': 'HOLIDROUGHT_SXI_31D',
    'version': (0, 1, 0),
    'path': '/p/data1/jibg31/poppe1/HOLIDROUGHT/SXI/31D/',
    'type_file': 'netcdf',
    'year_start': 1995,
    'month_start': 2,
    'year_end': 2018,
    'month_end': 12,
    'leapday': False,
    'resolution_time': '8D',
    'grid': 'EU3',
    'variables': ['SXI_GPP', 
                  'SXI_ESOIL',
                  'SXI_Tr',
                  'SXI_BTRAN',
                  'SXI_Gs',
                  'SXI_Runoff',
                  'SXI_SM_0',
                  'SXI_SM_1',
                  'SXI_SM_2',
                  'SXI_SM_3',
                  'SXI_SM_4',
                  'SXI_SM_5',
                  'SXI_SM_6',
                  'SXI_SM_7',
                  'SXI_SM_8',
                  'SXI_SM_9',
                  'SXI_SM_10',
                  'SXI_SM_11',
                  'SXI_SM_12',
                  'SXI_SM_13',
                  'SXI_SM_14',
                  'SXI_SM_15',
                  'SXI_SM_16',
                  'SXI_SM_17',
                  'SXI_SM_18',
                  'SXI_SM_19',
                  'SXI_WTD',
                  'SXI_VPD',
                  'SXI_P',
                  'SXI_SM_r'],
    'variable_names': {'SXI_GPP': 'SXI_GPP', 
                       'SXI_ESOIL': 'SXI_ESOIL',
                       'SXI_Tr': 'SXI_Tr',
                       'SXI_BTRAN': 'SXI_BTRAN',
                       'SXI_Gs': 'SXI_Gs',
                       'SXI_Runoff': 'SXI_Runoff',
                       'SXI_SM_0': 'SXI_SM_0',
                       'SXI_SM_1': 'SXI_SM_1',
                       'SXI_SM_2': 'SXI_SM_2',
                       'SXI_SM_3': 'SXI_SM_3',
                       'SXI_SM_4': 'SXI_SM_4',
                       'SXI_SM_5': 'SXI_SM_5',
                       'SXI_SM_6': 'SXI_SM_6',
                       'SXI_SM_7': 'SXI_SM_7',
                       'SXI_SM_8': 'SXI_SM_8',
                       'SXI_SM_9': 'SXI_SM_9',
                       'SXI_SM_10': 'SXI_SM_10',
                       'SXI_SM_11': 'SXI_SM_11',
                       'SXI_SM_12': 'SXI_SM_12',
                       'SXI_SM_13': 'SXI_SM_13',
                       'SXI_SM_14': 'SXI_SM_14',
                       'SXI_SM_15': 'SXI_SM_15',
                       'SXI_SM_16': 'SXI_SM_16',
                       'SXI_SM_17': 'SXI_SM_17',
                       'SXI_SM_18': 'SXI_SM_18',
                       'SXI_SM_19': 'SXI_SM_19',
                       'SXI_WTD': 'SXI_WTD',
                       'SXI_VPD': 'SXI_VPD',
                       'SXI_P': 'SXI_P',
                       'SXI_SM_r': 'SXI_SM_r'},
    'variable_dimensions': {'SXI_GPP': ['time', 'lat', 'lon'], 
                            'SXI_ESOIL': ['time', 'lat', 'lon'],
                            'SXI_Tr': ['time', 'lat', 'lon'],  
                            'SXI_BTRAN': ['time', 'lat', 'lon'],
                            'SXI_Gs': ['time', 'lat', 'lon'],
                            'SXI_Runoff': ['time', 'lat', 'lon'],
                            'SXI_SM_0': ['time', 'lat', 'lon'],
                            'SXI_SM_1': ['time', 'lat', 'lon'],
                            'SXI_SM_2': ['time', 'lat', 'lon'],
                            'SXI_SM_3': ['time', 'lat', 'lon'],
                            'SXI_SM_4': ['time', 'lat', 'lon'],
                            'SXI_SM_5': ['time', 'lat', 'lon'],
                            'SXI_SM_6': ['time', 'lat', 'lon'],
                            'SXI_SM_7': ['time', 'lat', 'lon'],
                            'SXI_SM_8': ['time', 'lat', 'lon'],
                            'SXI_SM_9': ['time', 'lat', 'lon'],
                            'SXI_SM_10': ['time', 'lat', 'lon'],
                            'SXI_SM_11': ['time', 'lat', 'lon'],
                            'SXI_SM_12': ['time', 'lat', 'lon'],
                            'SXI_SM_13': ['time', 'lat', 'lon'],
                            'SXI_SM_14': ['time', 'lat', 'lon'],
                            'SXI_SM_15': ['time', 'lat', 'lon'],
                            'SXI_SM_16': ['time', 'lat', 'lon'],
                            'SXI_SM_17': ['time', 'lat', 'lon'],
                            'SXI_SM_18': ['time', 'lat', 'lon'],
                            'SXI_SM_19': ['time', 'lat', 'lon'],
                            'SXI_WTD': ['time', 'lat', 'lon'],
                            'SXI_VPD': ['time', 'lat', 'lon'],
                            'SXI_P': ['time', 'lat', 'lon'],
                            'SXI_SM_r': ['time', 'lat', 'lon'],}, 
    'variable_units': {'SXI_GPP': 'dimensionless',
                       'SXI_ESOIL': 'dimensionless',
                       'SXI_Tr': 'dimensionless',
                       'SXI_BTRAN': 'dimensionless',
                       'SXI_Gs': 'dimensionless',
                       'SXI_Runoff': 'dimensionless',
                       'SXI_SM_0': 'dimensionless',
                       'SXI_SM_1': 'dimensionless',
                       'SXI_SM_2': 'dimensionless',
                       'SXI_SM_3': 'dimensionless',
                       'SXI_SM_4': 'dimensionless',
                       'SXI_SM_5': 'dimensionless',
                       'SXI_SM_6': 'dimensionless',
                       'SXI_SM_7': 'dimensionless',
                       'SXI_SM_8': 'dimensionless',
                       'SXI_SM_9': 'dimensionless',
                       'SXI_SM_10': 'dimensionless',
                       'SXI_SM_11': 'dimensionless',
                       'SXI_SM_12': 'dimensionless',
                       'SXI_SM_13': 'dimensionless',
                       'SXI_SM_14': 'dimensionless',
                       'SXI_SM_15': 'dimensionless',
                       'SXI_SM_16': 'dimensionless',
                       'SXI_SM_17': 'dimensionless',
                       'SXI_SM_18': 'dimensionless',
                       'SXI_SM_19': 'dimensionless',
                       'SXI_WTD': 'dimensionless',
                       'SXI_VPD': 'dimensionless',
                       'SXI_P': 'dimensionless',
                       'SXI_SM_r': 'dimensionless'},
    'mask_value': None
}


cluster_365D_EU3_8daily = {
    'name': 'HOLIDROUGHT_clusters_365D',
    'version': (0, 1, 0),
    'path': '/p/data1/jibg31/poppe1/HOLIDROUGHT/cluster/',
    'type_file': 'netcdf',
    'year_start': 1996,
    'month_start': 1,
    'year_end': 2018,
    'month_end': 12,
    'resolution_time': '8D',
    'grid': 'EU3',
    'leapday': False,
    'variables': ['SXI_GPP', 
                  'SXI_ESOIL',
                  'SXI_Tr',
                  'SXI_BTRAN',
                  'SXI_Gs',
                  'SXI_Runoff',
                  'SXI_SM',
                  'SXI_WTD',
                  'SXI_VPD',
                  'SXI_P',
                  'SXI_SM_r'],
    'variable_names': {'SXI_GPP': 'SXI_GPP', 
                       'SXI_ESOIL': 'SXI_ESOIL',
                       'SXI_Tr': 'SXI_Tr',
                       'SXI_BTRAN': 'SXI_BTRAN',
                       'SXI_Gs': 'SXI_Gs',
                       'SXI_Runoff': 'SXI_Runoff',
                       'SXI_SM': 'SXI_SM',
                       'SXI_WTD': 'SXI_WTD',
                       'SXI_VPD': 'SXI_VPD',
                       'SXI_P': 'SXI_P',
                       'SXI_SM_r': 'SXI_SM_r'},
    'variable_dimensions': {'SXI_GPP': ['time', 'lat', 'lon'], 
                            'SXI_ESOIL': ['time', 'lat', 'lon'],
                            'SXI_Tr': ['time', 'lat', 'lon'],  
                            'SXI_BTRAN': ['time', 'lat', 'lon'],
                            'SXI_Gs': ['time', 'lat', 'lon'],
                            'SXI_Runoff': ['time', 'lat', 'lon'],
                            'SXI_SM': ['time', 'lat', 'lon'],
                            'SXI_WTD': ['time', 'lat', 'lon'],
                            'SXI_VPD': ['time', 'lat', 'lon'],
                            'SXI_P': ['time', 'lat', 'lon'],
                            'SXI_SM_r': ['time', 'lat', 'lon'],}, 
    'variable_units': {'SXI_GPP': 'dimensionless',
                       'SXI_ESOIL': 'dimensionless',
                       'SXI_Tr': 'dimensionless',
                       'SXI_BTRAN': 'dimensionless',
                       'SXI_Gs': 'dimensionless',
                       'SXI_Runoff': 'dimensionless',
                       'SXI_SM': 'dimensionless',
                       'SXI_WTD': 'dimensionless',
                       'SXI_VPD': 'dimensionless',
                       'SXI_P': 'dimensionless',
                       'SXI_SM_r': 'dimensionless'},
    'mask_value': None,

    'agg_period': '365D',
    'grid_window': 3,
    'frac_points': 0.1,
    'weigh_mid': -1.4,
    'weigh_steep': 6.0,
    'filter_area': 30000.0,
    'factor_time': 1.0,
}

cluster_183D_EU3_8daily = {
    'name': 'HOLIDROUGHT_clusters_183D',
    'version': (0, 1, 0),
    'path': '/p/data1/jibg31/poppe1/HOLIDROUGHT/cluster/',
    'type_file': 'netcdf',
    'year_start': 1995,
    'month_start': 7,
    'year_end': 2018,
    'month_end': 12,
    'resolution_time': '8D',
    'grid': 'EU3',
    'leapday': False,
    'variables': ['SXI_GPP', 
                  'SXI_ESOIL',
                  'SXI_Tr',
                  'SXI_BTRAN',
                  'SXI_Gs',
                  'SXI_Runoff',
                  'SXI_SM',
                  'SXI_WTD',
                  'SXI_VPD',
                  'SXI_P',
                  'SXI_SM_r'],
    'variable_names': {'SXI_GPP': 'SXI_GPP', 
                       'SXI_ESOIL': 'SXI_ESOIL',
                       'SXI_Tr': 'SXI_Tr',
                       'SXI_BTRAN': 'SXI_BTRAN',
                       'SXI_Gs': 'SXI_Gs',
                       'SXI_Runoff': 'SXI_Runoff',
                       'SXI_SM': 'SXI_SM',
                       'SXI_WTD': 'SXI_WTD',
                       'SXI_VPD': 'SXI_VPD',
                       'SXI_P': 'SXI_P',
                       'SXI_SM_r': 'SXI_SM_r'},
    'variable_dimensions': {'SXI_GPP': ['time', 'lat', 'lon'], 
                            'SXI_ESOIL': ['time', 'lat', 'lon'],
                            'SXI_Tr': ['time', 'lat', 'lon'],  
                            'SXI_BTRAN': ['time', 'lat', 'lon'],
                            'SXI_Gs': ['time', 'lat', 'lon'],
                            'SXI_Runoff': ['time', 'lat', 'lon'],
                            'SXI_SM': ['time', 'lat', 'lon'],
                            'SXI_WTD': ['time', 'lat', 'lon'],
                            'SXI_VPD': ['time', 'lat', 'lon'],
                            'SXI_P': ['time', 'lat', 'lon'],
                            'SXI_SM_r': ['time', 'lat', 'lon'],}, 
    'variable_units': {'SXI_GPP': 'dimensionless',
                       'SXI_ESOIL': 'dimensionless',
                       'SXI_Tr': 'dimensionless',
                       'SXI_BTRAN': 'dimensionless',
                       'SXI_Gs': 'dimensionless',
                       'SXI_Runoff': 'dimensionless',
                       'SXI_SM': 'dimensionless',
                       'SXI_WTD': 'dimensionless',
                       'SXI_VPD': 'dimensionless',
                       'SXI_P': 'dimensionless',
                       'SXI_SM_r': 'dimensionless'},
    'mask_value': None,
    'agg_period': '183D',
    'grid_window': 3,
    'frac_points': 0.1,
    'weigh_mid': -1.4,
    'weigh_steep': 6.0,
    'filter_area': 30000.0,
    'factor_time': 1.0,
}

cluster_92D_EU3_8daily = {
    'name': 'HOLIDROUGHT_clusters_92D',
    'version': (0, 1, 0),
    'path': '/p/data1/jibg31/poppe1/HOLIDROUGHT/cluster/',
    'type_file': 'netcdf',
    'year_start': 1995,
    'month_start': 4,
    'year_end': 2018,
    'month_end': 12,
    'resolution_time': '8D',
    'grid': 'EU3',
    'leapday': False,
    'variables': ['SXI_GPP', 
                  'SXI_ESOIL',
                  'SXI_Tr',
                  'SXI_BTRAN',
                  'SXI_Gs',
                  'SXI_Runoff',
                  'SXI_SM',
                  'SXI_WTD',
                  'SXI_VPD',
                  'SXI_P',
                  'SXI_SM_r'],
    'variable_names': {'SXI_GPP': 'SXI_GPP', 
                       'SXI_ESOIL': 'SXI_ESOIL',
                       'SXI_Tr': 'SXI_Tr',
                       'SXI_BTRAN': 'SXI_BTRAN',
                       'SXI_Gs': 'SXI_Gs',
                       'SXI_Runoff': 'SXI_Runoff',
                       'SXI_SM': 'SXI_SM',
                       'SXI_WTD': 'SXI_WTD',
                       'SXI_VPD': 'SXI_VPD',
                       'SXI_P': 'SXI_P',
                       'SXI_SM_r': 'SXI_SM_r'},
    'variable_dimensions': {'SXI_GPP': ['time', 'lat', 'lon'], 
                            'SXI_ESOIL': ['time', 'lat', 'lon'],
                            'SXI_Tr': ['time', 'lat', 'lon'],  
                            'SXI_BTRAN': ['time', 'lat', 'lon'],
                            'SXI_Gs': ['time', 'lat', 'lon'],
                            'SXI_Runoff': ['time', 'lat', 'lon'],
                            'SXI_SM': ['time', 'lat', 'lon'],
                            'SXI_WTD': ['time', 'lat', 'lon'],
                            'SXI_VPD': ['time', 'lat', 'lon'],
                            'SXI_P': ['time', 'lat', 'lon'],
                            'SXI_SM_r': ['time', 'lat', 'lon'],}, 
    'variable_units': {'SXI_GPP': 'dimensionless',
                       'SXI_ESOIL': 'dimensionless',
                       'SXI_Tr': 'dimensionless',
                       'SXI_BTRAN': 'dimensionless',
                       'SXI_Gs': 'dimensionless',
                       'SXI_Runoff': 'dimensionless',
                       'SXI_SM': 'dimensionless',
                       'SXI_WTD': 'dimensionless',
                       'SXI_VPD': 'dimensionless',
                       'SXI_P': 'dimensionless',
                       'SXI_SM_r': 'dimensionless'},
    'mask_value': None,
    'agg_period': '92D',
    'grid_window': 3,
    'frac_points': 0.1,
    'weigh_mid': -1.4,
    'weigh_steep': 6.0,
    'filter_area': 30000.0,
    'factor_time': 1.0,
}

cluster_31D_EU3_8daily = {
    'name': 'HOLIDROUGHT_clusters_31D',
    'version': (0, 1, 0),
    'path': '/p/data1/jibg31/poppe1/HOLIDROUGHT/cluster/',
    'type_file': 'netcdf',
    'year_start': 1995,
    'month_start': 2,
    'year_end': 2018,
    'month_end': 12,
    'resolution_time': '8D',
    'grid': 'EU3',
    'leapday': False,
    'variables': ['SXI_GPP', 
                  'SXI_ESOIL',
                  'SXI_Tr',
                  'SXI_BTRAN',
                  'SXI_Gs',
                  'SXI_Runoff',
                  'SXI_SM',
                  'SXI_WTD',
                  'SXI_VPD',
                  'SXI_P',
                  'SXI_SM_r'],
    'variable_names': {'SXI_GPP': 'SXI_GPP', 
                       'SXI_ESOIL': 'SXI_ESOIL',
                       'SXI_Tr': 'SXI_Tr',
                       'SXI_BTRAN': 'SXI_BTRAN',
                       'SXI_Gs': 'SXI_Gs',
                       'SXI_Runoff': 'SXI_Runoff',
                       'SXI_SM': 'SXI_SM',
                       'SXI_WTD': 'SXI_WTD',
                       'SXI_VPD': 'SXI_VPD',
                       'SXI_P': 'SXI_P',
                       'SXI_SM_r': 'SXI_SM_r'},
    'variable_dimensions': {'SXI_GPP': ['time', 'lat', 'lon'], 
                            'SXI_ESOIL': ['time', 'lat', 'lon'],
                            'SXI_Tr': ['time', 'lat', 'lon'],  
                            'SXI_BTRAN': ['time', 'lat', 'lon'],
                            'SXI_Gs': ['time', 'lat', 'lon'],
                            'SXI_Runoff': ['time', 'lat', 'lon'],
                            'SXI_SM': ['time', 'lat', 'lon'],
                            'SXI_WTD': ['time', 'lat', 'lon'],
                            'SXI_VPD': ['time', 'lat', 'lon'],
                            'SXI_P': ['time', 'lat', 'lon'],
                            'SXI_SM_r': ['time', 'lat', 'lon'],}, 
    'variable_units': {'SXI_GPP': 'dimensionless',
                       'SXI_ESOIL': 'dimensionless',
                       'SXI_Tr': 'dimensionless',
                       'SXI_BTRAN': 'dimensionless',
                       'SXI_Gs': 'dimensionless',
                       'SXI_Runoff': 'dimensionless',
                       'SXI_SM': 'dimensionless',
                       'SXI_WTD': 'dimensionless',
                       'SXI_VPD': 'dimensionless',
                       'SXI_P': 'dimensionless',
                       'SXI_SM_r': 'dimensionless'},
    'mask_value': None,
    'agg_period': '31D',
    'grid_window': 3,
    'frac_points': 0.1,
    'weigh_mid': -1.4,
    'weigh_steep': 6.0,
    'filter_area': 30000.0,
    'factor_time': 1.0,
}

def create_yearly_files(path_rawdata: str,
                        file: str,
                        path_out: str,
                        time_index: pd.Series | pd.Index):
    
    """
    Input directory should only contain the GLEAM rawdata files...
    """
    
    create_dirs(path_out)

    files = glob(f'{path_rawdata}/{file}')

    data_raw = xr.open_mfdataset(files)
    
    data_raw = data_raw.assign_coords(time = time_index)

    years = np.unique(data_raw.time.dt.year.values)

    for y in years:

        print(f'Create yearly file for year {y}...\n')

        if check_file_exists(f'{path_out}/{y}.nc'): continue

        data_y = data_raw.sel(time = data_raw.time.dt.year.isin([y]))

        data_y.to_netcdf(f'{path_out}/{y}.nc',
                         format = 'NETCDF4_CLASSIC', 
                         unlimited_dims = ['time'])



EU3 = {
    'name': 'EUCORDEX3km',
    'version': (1, 0, 0),
    'path_file': '/p/scratch/cjibg31/jibg3105/CESMDataRoot/InputData/share/domains/',
    'name_file': 'EU3_CORDEX_gridfile_reducedim.nc',
    'type_file': 'netcdf',
    'name_latitude': 'XLAT_M',
    'name_longitude': 'XLONG_M',
    'name_landmask': 'LANDMASK',
    'lon_extents': [351.1, 57],
    'lat_extents': [27, 65.7],
}


GLEAM = {
    'name': 'GLEAM',
    'version': (3, 8, 1),
    'path_file': '/p/scratch/cjibg31/jibg3105/data/GLEAM/3.8a/yearly/',
    'name_file': '2000.nc',
    'type_file': 'netcdf',
    'name_latitude': 'lat',
    'name_longitude': 'lon',
    'lon_extents': [351.1, 57],
    'lat_extents': [27, 65.7],
}


ERA5L_EUCORDEX = {
    'name': 'ERA5-Land-EUCORDEX',
    'version': (1, 0, 0),
    'path_file': '/p/scratch/cjibg31/jibg3105/data/ERA5L/all/',
    'name_file': '2000.nc',
    'type_file': 'netcdf',
    'name_latitude': 'latitude',
    'name_longitude': 'longitude',
    'lon_extents': [351.1, 57],
    'lat_extents': [27, 65.7],
}

GLDAS_EUROCORDEX = {
    'name': 'GLDAS_EUROCORDEX',
    'version': (1, 0, 0),
    'path_file': '/p/scratch/cjibg31/jibg3105/data/GLDAS/NOAH025_M/',
    'name_file': '2000.nc',
    'type_file': 'netcdf',
    'name_latitude': 'lat',
    'name_longitude': 'lon',
    'lon_extents': [351.1, 57],
    'lat_extents': [27, 65.7],
}